#! /gpfs/user/menghaowei/anaconda2/bin/python
# _*_ coding: UTF-8 _*_


# Version information START --------------------------------------------------
VERSION_INFO = \
"""
Author: MENG Howard

Version-01:
  2019-05-24 解析mpileup文件；

Version-02:
  2019-06-04 增加多核运算功能

Version-03:
  2019-08-01 修复bug，当^X出现时会导致程序崩溃

Version-04:
  2019-08-02 修复bug，对insertion, deletion的信号能够准确识别

E-Mail: meng_howard@126.com
"""
# Version information END ----------------------------------------------------

# Learning Part START --------------------------------------------------------
LEARNING_PART = \
"""
Input samtools mpileup file
"""
# Learning Part END-----------------------------------------------------------

import argparse
import gzip
import sys
import os 
import string
import random
import time
from multiprocessing import Process

###############################################################################
# function part 
###############################################################################
def split_file_and_make_temp(input_filename, threads_num=1,temp_dir=None):
    """
    """
    #--------------------------------------------------->>>>>>
    # set temp dir 
    #--------------------------------------------------->>>>>>        
    if temp_dir is None:
        temp_dir = os.path.dirname(input_filename)
    
    #--------------------------------------------------->>>>>>
    # get input basename 
    #--------------------------------------------------->>>>>>    
    input_file_basename = os.path.basename(input_filename)

    #--------------------------------------------------->>>>>>
    # creat temp file name and open file 
    #--------------------------------------------------->>>>>>    
    #>>>>>> LOG 
    sys.stderr.write("Making temp files... \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))

    temp_filename_list = []
    temp_file_list = []
    
    for index in range(threads_num):
        # make filename 
        temp_file_basename = "temp_" + input_file_basename + "." + str(index) + "." +  "".join(random.sample(string.ascii_letters + string.digits, 16))
        temp_file_name = os.path.join(temp_dir , temp_file_basename)
        temp_filename_list.append(temp_file_name)
        
        # open file
        temp_file = open(temp_file_name, "w") 
        temp_file_list.append(temp_file)
    
    #--------------------------------------------------->>>>>>
    # counting input file line number
    #--------------------------------------------------->>>>>>   
    #>>>>>> LOG 
    sys.stderr.write("Counting input file... \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))
    
    total_input_line_num = 0
    
    if ".gz" == input_filename[-3:]:
        input_file = gzip.open(input_filename,"r")
    else:
        input_file = open(input_filename,"r")

    for line in input_file:
        total_input_line_num += 1 

    if (total_input_line_num % threads_num == 0):
        each_file_line_num = (total_input_line_num // threads_num)
    else:
        each_file_line_num = (total_input_line_num // threads_num) + 1

    input_file.close() 

    #>>>>>> LOG 
    sys.stderr.write("Done! \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))  

    #--------------------------------------------------->>>>>>
    # split temp files
    #--------------------------------------------------->>>>>> 
    # write into output filename 
    if ".gz" in input_filename:
        input_file = gzip.open(input_filename,"r")
    else:
        input_file = open(input_filename,"r")

    for index, line in enumerate(input_file):
        file_index = index // each_file_line_num 
        temp_file_list[file_index].write(line)

    # close output filename 
    input_file.close()
    [temp_file.close() for temp_file in temp_file_list]

    #>>>>>> LOG     
    sys.stderr.write("Make temp files done! \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))
    
    return(temp_filename_list)



def run_part(INPUT_FILE_PATH, OUTPUT_FILE_PATH="Stdout", MUT_NUM=0):
    """
    input_filename
    output_filename
    """
    
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # open input file
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    if ".gz" == INPUT_FILE_PATH[-3:]:
        INPUT_FILE = gzip.open(INPUT_FILE_PATH,"r")
    else:
        INPUT_FILE = open(INPUT_FILE_PATH,"r")

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # open output file
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    if OUTPUT_FILE_PATH == "Stdout":
        OUTPUT_FILE = sys.stdout
    elif ".gz" == OUTPUT_FILE_PATH[-3:]:
        OUTPUT_FILE = gzip.open(OUTPUT_FILE_PATH,"w")
    else:
        OUTPUT_FILE = open(OUTPUT_FILE_PATH,"w")

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # main part
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # parse mpileup file
    for line in INPUT_FILE:
        data = line.strip().split('\t')
        chr_name = data[0]        
        bp = data[1]
        bases = data[4].upper()
        ref = data[2].upper()

        types = {'A':0,'G':0,'C':0,'T':0,'-':[],'+':[],'Not':[]}

        # coverage > 0 
        i = 0
        while i < len(bases):
            base = bases[i]

            if base == '^' :
                i += 2
            
            elif base == '$':
                i += 1 

            elif base == '-':
                i += 1
                del_str_num = ""
                
                while bases[i].isdigit():
                    del_str_num += bases[i]
                    i += 1
                
                delNum = int(del_str_num)
                delSeq = ""
                for a in range(delNum):
                    delSeq += bases[i]
                    i += 1
                types['-'].append(delSeq)

            elif base == '*':
                types['-'].append(bases[i])
                i += 1

            elif base == '+':
                i += 1
                add_str_num = ""
                
                while bases[i].isdigit():
                    add_str_num += bases[i]
                    i += 1
                    
                addNum = int(add_str_num)
                addSeq = ""
                for a in range(addNum):
                    addSeq += bases[i]
                    i += 1
                types['+'].append(addSeq)
                
            elif (base == '.') or (base == ','):
                types[ref] += 1
                i += 1

            else:
                if types.has_key(base):
                    types[base] += 1
                else:
                    types['Not'].append(base)
                i += 1

        adds = '.'
        adds_count = 0
        if len(types['+']) > 0:
            adds = ','.join(types['+'])
            adds_count = len(types['+'])
            
        dels = "."
        dels_count = 0
        if len(types['-']) > 0:
            dels = ",".join(types['-'])
            dels_count = len(types['-'])

        amb = '.'
        amb_count = 0
        if len(types['Not']) > 0:
            amb = ','.join(types['Not'])
            amb_count = len(types['Not'])

        # get other mutation number 
        if types.has_key(ref):
            mut_num = types["A"] + types["T"] + types["G"] + types["C"] - types[ref]
        else:
            mut_num = 0

        out_list = [
            chr_name,
            bp,
            ref,
            types['A'],
            types['G'],
            types['C'],
            types['T'],
            dels_count,
            adds_count,
            amb_count,
            dels,
            adds,
            amb,
            mut_num
        ]

        # make a filter 
        if mut_num >= MUT_NUM:
            OUTPUT_FILE.write("\t".join(map(str,out_list)) + "\n")


    INPUT_FILE.close()
    OUTPUT_FILE.close()


def merge_out_bmat(temp_filename_list, output_filename, remove_temp=True):
    """
    """
    # open output file
    if output_filename == "Stdout":
        output_file = sys.stdout
    elif ".gz" == OUTPUT_FILE_PATH[-3:]:
        output_file = gzip.open(output_filename,"w")
    else:
        output_file = open(output_filename,"w")
    
    # write header  
    header = [
        "chr_name",
        "chr_index",
        "ref_base",
        "A",
        "G",
        "C",
        "T",
        "del_count",
        "insert_count",
        "ambiguous_count",
        "deletion",
        "insertion",
        "ambiguous",
        "mut_num"
    ]
    output_file.write("\t".join(header) + "\n")    


    # open temp bmat file 
    for temp_filename in temp_filename_list:
        # write file 
        with open(temp_filename + ".bmat","r") as temp_file:
            for line in temp_file:
                output_file.write(line)
    
    # remove temp files
    if remove_temp:
        for temp_filename in temp_filename_list:
            os.remove(temp_filename)
            os.remove(temp_filename + ".bmat")

###############################################################################
# read parameters
###############################################################################
if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="convert mpileup file to info file")

    parser.add_argument("-i", "--Input",
        help="samtools mpileup format file",required=True)

    parser.add_argument("-o", "--Output",
        help="Output parsed file",default="Stdout")

    parser.add_argument("-p", "--Threads",
        help="Multiple threads number, default=1",default="1")

    parser.add_argument("-n", "--MutNum",
        help="Only contain mutation info go to the output, set 0 mean output all site, default=0",default="0")

    parser.add_argument("--TempDir",
        help="Where to keep temp files, default is the same dir with --Input",default=None)


    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # load the paramters
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    ARGS = parser.parse_args()

    INPUT_FILE_PATH = ARGS.Input
    OUTPUT_FILE_PATH = ARGS.Output

    THREADS_NUM = int(ARGS.Threads)
    MUT_NUM = int(ARGS.MutNum)

    TEMP_DIR = ARGS.TempDir

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # make temp files
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    temp_filename_list = split_file_and_make_temp(input_filename=INPUT_FILE_PATH, threads_num=THREADS_NUM,temp_dir=TEMP_DIR)
    
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # multiple processing part 
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    sys.stderr.write("Parsing files... \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))

    procs_list = []

    for index,temp_filename in enumerate(temp_filename_list):
        out_filename =  temp_filename + ".bmat"
        
        # add multiprocess
        sub_proc = Process(target=run_part, args=(temp_filename,out_filename,MUT_NUM,))
        procs_list.append(sub_proc)
        sub_proc.start()

    for sub_proc in procs_list:
        sub_proc.join()

    sys.stderr.write("Done! \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))        

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # merge output files
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    #>>>>>> LOG  
    sys.stderr.write("Merging files... \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))

    merge_out_bmat(temp_filename_list=temp_filename_list, output_filename=OUTPUT_FILE_PATH, remove_temp=True)

    #>>>>>> LOG  
    sys.stderr.write("Done!... \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())))



    

    









# 2019-08-02


from bioat._version import version as __version__
from bioat._version import author as __author__
from bioat._version import email as __email__
from pysam import __version__ as __pysam_version__
from pysamstats import __version__ as __pysamstats_version__


import logging
import math
import os
import sys

import numpy as np
import pandas as pd

"""initialize"""
pd.set_option("max_colwidth", 60)  # column最大宽度
pd.set_option("display.width", 200)  # dataframe宽度
pd.set_option("display.max_columns", None)  # column最大显示数
pd.set_option("display.max_rows", 100)  # row最大显示数


def logging_setting():
    # log setting
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)-5s @ %(asctime)s: %(message)s ",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        filemode="w",
        force=True,
    )


if __name__ == "__main__":
    # logging_setting
    logging_setting()
    # file list path
    pt_target = "../additional_file/20220104_DdCBE-target-seq-info.csv,"
    pt_detect = "../additional_file/20220104_DdCBE-detect-seq-info_ND4-Det_rep1.csv, ../additional_file/20220104_DdCBE-detect-seq-info_ND4-Det_rep2.csv,../additional_file/20220104_DdCBE-detect-seq-info_ND5-1-Det_rep1.csv,../additional_file/20220104_DdCBE-detect-seq-info_ND5-1-Det_rep2.csv,../additional_file/20220104_DdCBE-detect-seq-info_ND6-Det_rep1.csv,../additional_file/20220104_DdCBE-detect-seq-info_ND6-Det_rep2.csv"
    pt_out = "../additional_file/20220104_DdCBE-Merged-info.csv"
    # deal with file list
    ls_target = [i.strip() for i in pt_target.split(",") if i.strip() != ""]
    ls_detect = [i.strip() for i in pt_detect.split(",") if i.strip() != ""]

    df_target = pd.DataFrame()

    for pt_df in ls_target:
        df = pd.read_csv(pt_df, sep=",", header=0, index_col=False)
        df_target = pd.concat([df_target, df], axis=0, ignore_index=True)

    df_detect = pd.DataFrame()

    for pt_df in ls_detect:
        df = pd.read_csv(pt_df, sep=",", header=0, index_col=False, dtype={'A_ratio': np.float64, 'T_ratio': np.float64, 'C_ratio': np.float64, 'G_ratio': np.float64})
        df_detect = pd.concat([df_detect, df], axis=0, ignore_index=True)

    # for test!!
    # "".join(
    #     df_target[
    #         (df_target.treatment == "N6-2000-12")
    #         & (df_target.rep == "rep1")
    #         & (df_target.cutoff == 3)
    #         & (df_target.region_id == "ND6-only-1")
    #         & (df_target.relative_pos >= 15)
    #         & (df_target.mut_base == "C")
    #     ].ref_base
    # )

    # "".join(
    #     df_detect[
    #         (df_detect.treatment == "ND6-Det")
    #         & (df_detect.rep == "rep1")
    #         & (df_detect.region_id == "ND6-only-1")
    #         & (df_detect.relative_pos >= 15)
    #     ].genome_base
    # )
    # fix target rep3 to detect rep1
    df_target["rep_back_target"] = pd.Series(dtype=object)
    df_target.rep_back_target = df_target.rep
    df_target.loc[df_target.rep == "rep3", "rep"] = "rep1"
    df_target[df_target.rep_back_target == "rep3"]
    # fix na
    df_detect.fillna(0)
    for col in df_detect[df_detect.A_ratio.map(math.isnan)][
        ["A_ratio", "G_ratio", "C_ratio", "T_ratio"]
    ]:
        df_detect.loc[:, col] = 0
    df_detect.isna().sum()
    # merge detect-seq and target-seq
    df = pd.merge(
        left=df_target,
        right=df_detect,
        how="inner",
        on=["rep", "region_id", "relative_pos"],
        suffixes=("_TargetSeq", "_DetectSeq"),
        indicator=True,
    )
    # check na
    ifna = df.isna().sum().sum()

    if ifna != 0:
        logging.warn("{} NA values were found in merged table!".format(ifna))

    logging.info("Merging done!")
    df
    df.to_csv(pt_out, sep=",", index=False, header=True)



import argparse
import gzip
import logging
import os
import sys

import numpy as np
import pandas as pd
from Bio import SeqIO

logging.basicConfig(
    level=(4 - 30) * 10,
    format="%(levelname)-5s @ %(asctime)s: %(message)s ",
    datefmt="%Y-%m-%d %H:%M:%S",
    stream=sys.stderr,
    filemode="w",
)


class SiteIndex(object):
    """SiteIndex.
    SiteIndex object can parse and return base mutation info.

    Args
        site_index (str): Like chr1_10000_CT or chr1_10000.

    Methods
        no methods

    Attributes
        chr_name (str): chr_name.
        chr_index (int): chr_index.
        site_index (str): like chr1_10000.
        site_index_raw (str): like chr1_10000_CT.

    Raises
        no raises.
    """

    def __init__(self, site_index):
        site_index_split = site_index.split("_")

        self.chr_name = site_index_split[0]
        self.chr_index = int(site_index_split[1])
        self.site_index = site_index_split[0] + "_" + site_index_split[1]
        self.site_index_raw = site_index

        if len(site_index_split) > 2:
            self.mut_type = site_index_split[2]


class BmatLine(object):
    """BmatLine.
    SiteIndex object can parse and return base mutation info.

    Args
        bmat_line_list (list): Like this:
            [
               'chr1',
               '10013',
               'T',
               '0',
               '0',
               '0',
               '6',
               '0',
               '0',
               '0',
               '.',
               '.',
               '.',
               '0'
            ]

    Methods
        no methods

    Attributes
        chr_name (str): chr_name.
        chr_index (int): chr_index.
        ref_base (str): ref_base.
        count_del (int): count_del.
        count_insert (int): count_insert.
        count_ambis (int): count_ambis.
        deletion (str): deletion.
        ambiguous (str): ambiguous.
        insertion (str): insertion.
        mut_num (int): mut_num.
        count_dict (dict): like this:
            {
                'A': 0,
                'G': 0,
                'C': 10,
                'T': 25,
            }.

    Raises
        no raises.
    """

    def __init__(self, bmat_line_list):
        self.chr_name = bmat_line_list[0]
        self.chr_index = int(bmat_line_list[1])
        self.ref_base = bmat_line_list[2]

        self.count_del = int(bmat_line_list[7])
        self.count_insert = int(bmat_line_list[8])
        self.count_ambis = int(bmat_line_list[9])

        self.deletion = bmat_line_list[10]
        self.ambiguous = bmat_line_list[11]
        self.insertion = bmat_line_list[12]
        self.mut_num = int(bmat_line_list[13])

        # make dict
        self.count_dict = {
            "A": int(bmat_line_list[3]),
            "G": int(bmat_line_list[4]),
            "C": int(bmat_line_list[5]),
            "T": int(bmat_line_list[6]),
        }


def parser():
    """CLI interface.

    Args
        no args.

    Returns
        argparse.Namespace
    Raises
        no raises.
    """
    args = argparse.ArgumentParser(
        description="A tools for formating MPMAT file from BED6/BED12 file with necessary fake information."
    )
    args.add_argument(
        "--bedHasHeader",
        type=bool,
        required=True,
        help="True if bed has header, " "else False",
    )
    args.add_argument(
        "--bmatHasHeader",
        type=bool,
        required=True,
        help="True if bmat has header, else False",
    )
    args.add_argument("-b", "--bed", required=True, help="Path of the bed file")
    args.add_argument(
        "-m",
        "--bmat",
        required=True,
        help="Path of the bmat file, gzip file is accepted",
    )
    args.add_argument(
        "-g", "--genome", required=True, help="Path of the genome fasta file"
    )
    args.add_argument(
        "-o",
        "--output_mpmat",
        default="./output.mpmat",
        help="Path of the fake mpmat file, default: ./output.mpmat",
    )
    args.add_argument(
        "--extend_length_PAM_term",
        type=int,
        default=0,
        help="Pextend length at PAM term, default: int 0",
    )
    args.add_argument(
        "--extend_length_far_end_term_from_PAM",
        type=int,
        default=0,
        help="Pextend length at far end term from PAM, default: int 0",
    )
    parser = args.parse_args()

    return parser


def cmp_site_index(site_index_a, site_index_b, ref_order_dict):
    """cmp_site_index.

    Args
        site_index_a (str): like chr1_629627_CT.
        site_index_b (str): like chr1_629627_CT.
        ref_order_dict (dict): like this:
            {
                'chr1': 0,
                'chr19': 1,
                'chr20': 2
            }

    Returns
        status (int): -1|0|1, details:
            -1, site_a at upstream of site_b
            0, site_index_a == site_index_b at position level, ignore mutation info
            1, site_a at downstream of site_b

    Raises
        TypeError: see codes for more.
    """
    # 这里可能有个问题就是有ref 没chr， 有ref 有chr
    # 比如 chr1_939943_CT 1_10006_C这两个去比较
    # 策略是加上chr，肯定不够通用，以后再说改的事儿
    if not site_index_a.startswith("chr"):
        site_index_a = f"chr{site_index_a}"

    if not site_index_b.startswith("chr"):
        site_index_b = f"chr{site_index_b}"

    site_A = SiteIndex(site_index_a)
    site_B = SiteIndex(site_index_b)

    if site_A.chr_name == site_B.chr_name:
        if site_A.chr_index == site_B.chr_index:
            return 0

        elif site_A.chr_index < site_B.chr_index:
            return -1

        elif site_A.chr_index > site_B.chr_index:
            return 1

    else:
        site_A_chr_order = ref_order_dict.get(site_A.chr_name)
        site_B_chr_order = ref_order_dict.get(site_B.chr_name)

        if (site_A_chr_order is not None) and (site_B_chr_order is not None):
            if site_A_chr_order < site_B_chr_order:
                return -1

            elif site_A_chr_order > site_B_chr_order:
                return 1

        else:
            raise TypeError("Site index not in your reference!")


def load_reference_fasta_as_dict(ref_fasta_path, ref_name_list="All", log_verbose=30):
    """load_reference_fasta_as_dict.

    Args
        ref_fasta_path (str): Reference fasta file path.
        ref_name_list (str): If set All, load all seq info in reference,
            else only try to load seq_id in the list.
        ref_order_dict (dict): like this:
            {
                'chr1': 0,
                'chr19': 1,
                'chr20': 2
            }

    Returns
        ref_seq_dict (dict): A dict, key is seq_id and value is sequence with .upper()
        or
        None (NoneType): If occur error, return None.

    Raises
        IOError: see codes for more.
    """
    # ---------------------------------------------------------->>>>>>>>>>>>>>>>>>>>
    # log setting
    # ---------------------------------------------------------->>>>>>>>>>>>>>>>>>>>
    logging.basicConfig(
        level=(4 - log_verbose) * 10,
        format="%(levelname)-5s @ %(asctime)s: %(message)s ",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        filemode="w",
    )

    # ---------------------------------------------------------->>>>>>>>>>>>>>>>>>>>
    # load genome as dict
    # ---------------------------------------------------------->>>>>>>>>>>>>>>>>>>>

    try:
        if ref_fasta_path.endswith(".gz"):
            genome_fa = SeqIO.parse(
                handle=gzip.open(ref_fasta_path, "rt"), format="fasta"
            )
        else:
            genome_fa = SeqIO.parse(handle=open(ref_fasta_path, "rt"), format="fasta")
    except:
        raise IOError("Load file error! %s" % ref_fasta_path)

    # init var
    ref_seq_dict = {}
    ref_name_set = set(ref_name_list)

    logging.info("Starting to load the reference genome...")

    for ref in genome_fa:
        if ref_name_list == "All":
            ref_seq_dict[ref.id] = ref.seq.upper()
            logging.debug("Loading genome...\t" + ref.id)

        elif ref.id in ref_name_list:
            ref_seq_dict[ref.id] = ref.seq.upper()
            logging.debug("Loading genome...\t" + ref.id)

            # remove already loaded seq
            ref_name_set.remove(ref.id)

            # load all info
            if len(ref_name_set) == 0:
                break
        # for debug
        # break

    del genome_fa
    logging.info("Loading genome done!")

    return ref_seq_dict


def query_region_bmat_info(bmat_file, site_index_list, genome_order_dict):
    """query_region_bmat_info.

    Args
        bmat_file (BmatLine object): Reference fasta file path.
        site_index_list (list): like [chr1_20452_CT, chr1_20467_C., chr1_20474_CT].
        genome_order_dict (dict): or FUN <cmp_site_index>, like this:
            {
                'chr1': 0,
                'chr19': 1,
                'chr20': 2
            }

    Returns
        site_dict (dict): ey is site_index, value bmat line list.

    Raises
        IOError: see codes for more.
    """
    site_index = SiteIndex(site_index_list[0])
    bmat_line = bmat_file.readline()
    if type(bmat_line) == str:
        pass
    else:
        bmat_line = bmat_line.decode(encoding="utf8")

    # define dict
    query_total_num = len(site_index_list)
    query_site_num = 0
    query_res_dict = {"site_index_list": []}

    while bmat_line != "":

        bmat_line_list = bmat_line.strip().split("\t")
        bmat_site_index = "_".join(bmat_line_list[0:3])

        cmp_res = cmp_site_index(
            site_index.site_index_raw, bmat_site_index, genome_order_dict
        )

        if cmp_res == 0:
            query_site_num += 1

            # add info into dict
            query_res_dict["site_index_list"].append(site_index.site_index)
            query_res_dict[site_index.site_index] = BmatLine(bmat_line_list)

            if query_site_num >= query_total_num:
                break

            else:
                # read new
                site_index = SiteIndex(site_index_list[query_site_num])
                bmat_line = bmat_file.readline()
                if type(bmat_line) == str:
                    pass
                else:
                    bmat_line = bmat_line.decode(encoding="utf8")

        elif cmp_res == 1:
            bmat_line = bmat_file.readline()
            if type(bmat_line) == str:
                pass
            else:
                bmat_line = bmat_line.decode(encoding="utf8")

        elif cmp_res == -1:
            query_site_num += 1

            # add info into dict
            query_res_dict["site_index_list"].append(site_index.site_index)
            query_res_dict[site_index.site_index] = None

            if query_site_num >= query_total_num:
                break

            else:
                # read new
                site_index = SiteIndex(site_index_list[query_site_num])
                bmat_line = bmat_file.readline()
                if type(bmat_line) == str:
                    pass
                else:
                    bmat_line = bmat_line.decode(encoding="utf8")

    return query_res_dict


if __name__ == "__main__":
    DEBUG = True
    if DEBUG:
        BED_PATH = "../Test_Resources/test.bed.gz"
        # BMAT_PATH = "../Test_Resources/test.bmat"
        BMAT_PATH = "../Test_Resources/test.bmat.gz"
        REF_PATH = "/Users/zhaohuanan/igv/genomes/seq/hg38.fa.gz"
        # mm10.fa.gz
        bmatHasHeader = True
        bedHasHeader = True
        OUT_MPMAT = "../Test_Outfiles/test.mpmat.gz"
        farTermExtend = 0
        pamTermExtend = 0
    else:
        # parse params
        parser = parser()
        BED_PATH = parser.bed
        BMAT_PATH = parser.bmat
        REF_PATH = parser.genome
        bmatHasHeader = parser.bmatHasHeader
        bedHasHeader = parser.bedHasHeader
        OUT_MPMAT = parser.output_mpmat
        farTermExtend = parser.extend_length_far_end_term_from_PAM
        pamTermExtend = parser.extend_length_PAM_term

    REF_FA = load_reference_fasta_as_dict(ref_fasta_path=REF_PATH, log_verbose=30)
    bmat_file = (
        gzip.open(BMAT_PATH, "rt")
        if BMAT_PATH.endswith(".gz")
        else open(BMAT_PATH, "rt")
    )
    # create output mpmat
    out_mpmat = (
        gzip.open(OUT_MPMAT, "at")
        if OUT_MPMAT.endswith(".gz")
        else open(OUT_MPMAT, "at")
    )

    if bmatHasHeader:
        fake_var = next(bmat_file)

    # 判断bed是否是空的, 空的抛出错误
    assert os.path.getsize(BED_PATH)

    # 将bed读作Dataframe
    if bedHasHeader:
        df = pd.read_csv(BED_PATH, sep="\t")
    else:
        df = pd.read_csv(BED_PATH, sep="\t", header=None)
    # 只取bed的前6列信息：BED6
    df = df.iloc[:, 0:7]
    df.columns = ["chrom", "start", "end", "name", "score", "strand"]
    # 将每个 染色体 内的 bed信息 提取至词典中 供后续 按照 染色体 调用
    dt_chrom_bed_info = {}

    for chrom, df_ in df.groupby("chrom"):
        df_ = df_.sort_values(by=["start"])
        df_.index = range(df_.shape[0])
        dt_chrom_bed_info[chrom] = df_

    for chrom, df_ in df.groupby("chrom"):
        if dt_chrom_bed_info[chrom].empty:
            # 如果 词典中，对应染色体的那部分bed是空的，就跳过，继续分析下一个染色体中的bed
            continue
        else:
            # 对于某特定 染色体 中的bed
            for row in range(dt_chrom_bed_info[chrom].shape[0]):
                # 遍历该Dataframe，每次取一行bed信息
                # 并指定 start、end、name、score、strand
                # 已知这一个bed区块的chrom信息
                # 则指定了chrom、start、end、name、score、strand
                start = dt_chrom_bed_info[chrom].loc[row, "start"]
                end = dt_chrom_bed_info[chrom].loc[row, "end"]
                name = dt_chrom_bed_info[chrom].loc[row, "name"]
                score = dt_chrom_bed_info[chrom].loc[row, "score"]
                strand = dt_chrom_bed_info[chrom].loc[row, "strand"]
                # seq and abs_index of bases
                # 从REF_FA的dict中将对应这一行bed信息的region的seq切片出来
                try:
                    seq = REF_FA[chrom][(start - 1) : end]
                except KeyError:
                    logging.warning(f"Skip Bed Coordinate: {chrom}\t{start}\t{end}")
                    continue

                seq = seq.upper()
                # 生成seq对应的染色体位置的绝对index
                seq_idx = np.arange(start, end + 1)
                # form <site_index_list>
                # It's a list
                # like [chr1_20452_CT, chr1_20467_C., chr1_20474_CT]
                # C don't mut to T --> C.
                # C mut to T --> CT

                site_index_list = []
                for idx, base in enumerate(seq):
                    # 调整start和end的值
                    # region为正向，就看正链REF中C的突变情况
                    # region为反向，就看正链REF中G的突变情况，对应到反向链，其实就还是看C
                    if strand == "+" and base == "C":
                        base2base = "CT"
                        start -= farTermExtend
                        end += pamTermExtend
                    elif strand == "-" and base == "G":
                        base2base = "GA"
                        start -= pamTermExtend
                        end += farTermExtend
                    else:
                        continue
                    site_index_list.append(
                        "{chrom}_{abs_idx}_{base2base}".format(
                            chrom=chrom,
                            abs_idx=seq_idx[idx],
                            base2base=base2base,
                        )
                    )
                if len(site_index_list) != 0 and site_index_list != [""]:
                    query_mut_info = query_region_bmat_info(
                        bmat_file=bmat_file,
                        site_index_list=site_index_list,
                        genome_order_dict=REF_FA,
                    )
                    # logging.debug("=" * 20)
                    # seqstr = "this seq: ", site_index_list
                    # logging.debug(seqstr)
                    site_index_list_mut_count = []
                    site_index_list_coverage = []
                    for base in site_index_list:
                        # basestr = "this base:" + base
                        # logging.debug(basestr)
                        try:
                            dt_this_base = query_mut_info[base[:-3]].count_dict
                            # logging.debug(dt_this_base)
                            site_index_list_mut_count.append(
                                int(dt_this_base[base[-1]])
                            )
                            coverage = 0
                            for key in dt_this_base:
                                coverage += int(dt_this_base[key])
                            site_index_list_coverage.append(int(coverage))
                        except KeyError:
                            site_index_list_mut_count.append(0)
                            site_index_list_coverage.append(0)

                    # logging.debug("=" * 20)
                    ls_ratio = []
                    for idx in range(len(site_index_list)):
                        if site_index_list_coverage[idx] == 0:
                            ls_ratio.append(0.0)
                        else:
                            ls_ratio.append(
                                site_index_list_mut_count[idx]
                                / (site_index_list_coverage[idx])
                            )
                    # logging.debug(ls_ratio)

                    count_mut_site_in_tandom = len(site_index_list)
                    for i in range(len(site_index_list_mut_count)):
                        if site_index_list_mut_count[i] == 0:
                            site_index_list[i] = site_index_list[i][:-1] + "."
                            count_mut_site_in_tandom -= 1

                    mpmat_line = (
                        "{chrom}\t{start_tandom}\t{end_tandom}\t{count_tandom_site}\t".format(
                            chrom=chrom,
                            start_tandom=site_index_list[0].split("_")[1],
                            end_tandom=site_index_list[-1].split("_")[1],
                            count_tandom_site=len(site_index_list),
                        )
                        + "{count_mut_site_in_tandom}\t{count_SNP_site_in_tandom}\t".format(
                            count_mut_site_in_tandom=count_mut_site_in_tandom,
                            count_SNP_site_in_tandom=0,
                        )
                        + "{mut_site_index}\t{mut_count_this_site}\t{mut_coverage_this_site}\t".format(
                            mut_site_index=",".join(site_index_list),
                            mut_count_this_site=",".join(
                                [str(i) for i in site_index_list_mut_count]
                            ),
                            mut_coverage_this_site=",".join(
                                [str(i) for i in site_index_list_coverage]
                            ),
                        )
                        + "{mut_ratio_this_site}\t".format(
                            mut_ratio_this_site=",".join([str(i) for i in ls_ratio])
                        )
                        + "{isSNP}\t{zeros}\t{passTest}\n".format(
                            isSNP=",".join(["False"] * len(site_index_list)),
                            zeros=",".join(["0"] * len(site_index_list)),
                            passTest=",".join(["Pass"] * len(site_index_list)),
                        )
                    )
                    # logging.debug("DEBUG: " + mpmat_line)
                    out_mpmat.write(mpmat_line)
                else:
                    # 如果seq中没有要check的base，比如+没C或-没G就pass这个点
                    pass

logging.info("The program done!")

bmat_file.close()
out_mpmat.close()



from cleo.commands.command import Command
from cleo.helpers import argument, option


class BedToolsCommand(Command):
    name = "bedtools"
    description = "Toolkit for bed file"
    arguments = [
        argument(
            "sort",
            description="Sort a bed file",
            optional=True
        )
    ]
    options = [
        option(
            long_name="--input",
            short_name="-i",
            description="Input bed file",
            flag=True,
            value_required=True,
            multiple=False,
            default=None,
        )
    ]

    def handle(self):
        name = self.argument("sort")
        self.line(name)




import argparse
import logging
import os
import sys

import numpy as np
import pandas as pd

"""initialize"""
pd.set_option("max_colwidth", 60)  # column最大宽度
pd.set_option("display.width", 150)  # dataframe宽度
pd.set_option("display.max_columns", None)  # column最大显示数
pd.set_option("display.max_rows", 100)  # row最大显示数

dt_type = {
    "chr_name": str,
    "chr_index": str,
    "ref_base": str,
    "A": int,
    "G": int,
    "C": int,
    "T": int,
    "del_count": int,
    "insert_count": int,
    "ambiguous_count": int,
    "deletion": str,
    "insertion": str,
    "ambiguous": str,
    "mut_num": int,
}


def logging_setting():
    # log setting
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)-5s @ %(asctime)s: %(message)s ",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        filemode="w",
        force=True,
    )


def argparse_setting():
    # argparse setting
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--bmat_folder",
        type=str,
        required=True,
        help="""
            All the bmat file should be formated as:
            TargetSeq-TREATMENT_x_REP_x_REGIONID_x_CUTOFF_x.bmat.gz
            and put in this folder.
        """,
    )
    parser.add_argument(
        "--out",
        type=str,
        default=None,
        help="""
            The path to put out csv file.
        """,
    )
    return parser.parse_args()


def return_base_ratio(x):
    if x["ref_base"] == "C":
        return x["T"] / x["depth"]
    elif x["ref_base"] == "G":
        return x["A"] / x["depth"]
    else:
        return np.NaN


if __name__ == "__main__":
    # logging_setting
    logging_setting()
    # argparse_setting
    args = argparse_setting()
    pt_bmat = args.bmat_folder
    pt_out = args.out
    ls_bmat = [os.path.join(pt_bmat, i) for i in os.listdir(pt_bmat) if "bmat" in i]
    # # TargetSeq-TREATMENT_x_REP_x_REGIONID_x_CUTOFF_x.bmat.gz

    # """ load bmat and merge """
    df_merge = pd.DataFrame()
    for path in ls_bmat:
        df = pd.read_csv(path, header=0, index_col=False, sep="\t", dtype=dt_type)
        df["bmat_name"] = path.replace("bmat/", "")
        df_merge = pd.concat([df_merge, df], ignore_index=True)

    # fix region_id for ND4 ND5.1 ND6
    # df_merge.chr_name = df_merge.chr_name.map(lambda x: x.replace("SHARE", "share"))
    # df_merge.chr_name = df_merge.chr_name.map(lambda x: x.replace("ONLY", "only"))
    # """ fix rep info """
    # remove no need cols
    df = df_merge[
        [
            "chr_name",
            "chr_index",
            "ref_base",
            "A",
            "G",
            "C",
            "T",
            "mut_num",
            "bmat_name",
        ]
    ].copy()

    # form rep info
    df["rep"] = df.bmat_name.map(
        lambda x: "rep" + x.split("_REP_")[1].split("_CUTOFF_")[0]
    )
    # form treatment info
    df["treatment"] = df.bmat_name.map(
        lambda x: x.split("-TREATMENT_")[1].split("_REP_")[0]
    )
    # form cutoff info
    df["cutoff"] = df.bmat_name.map(
        lambda x: int(x.split("_CUTOFF_")[1].split("_REGIONID_")[0])
    )

    if df.isna().sum().sum() != 0:
        raise ValueError("bmat file name should be formated")

    # select
    df_select = df[
        [
            "chr_name",
            "chr_index",
            "ref_base",
            "A",
            "G",
            "C",
            "T",
            "bmat_name",
            "rep",
            "treatment",
            "cutoff",
        ]
    ].copy()

    # df_select
    df_select["depth"] = df_select[["A", "G", "C", "T"]].apply(sum, axis=1)
    # 防止分母为0
    df_select.depth = df_select.depth.map(lambda x: x if x != 0 else 1)

    df_bmat_melt = df_select.melt(
        id_vars=[
            "treatment",
            "rep",
            "cutoff",
            "chr_name",
            "ref_base",
            "chr_index",
            "bmat_name",
            "depth",
        ],
        value_vars=["A", "G", "C", "T"],
        var_name="mut_base",
        value_name="mut_count",
    )

    df_bmat_melt.columns = [
        "treatment",
        "rep",
        "cutoff",
        "region_id",
        "ref_base",
        "relative_pos",
        "bmat_name",
        "total_count",
        "mut_base",
        "mut_count",
    ]
    df_bmat_melt.to_csv(pt_out, index=False, header=True)
    logging.info("script done.")



import argparse
import sys

from pyfasta import Fasta


def get_opt():
    """Check and parsing the opts"""
    parser = argparse.ArgumentParser(
        prog="seqFasta",
        description="seqFasta: A program to get fasta file from BED file ",
        usage="%(prog)s.py [options] -g genome -b bedfile",
    )
    parser.add_argument(
        "-g",
        "--genome",
        nargs="?",
        type=str,
        default=sys.stdin,
        help="The path of fasta genome file, the suffix must be '.fa' or '.fasta'. [Required]",
        required=True,
    )
    parser.add_argument(
        "-b",
        "--bedfile",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="BED6 file. detail, http://genome.ucsc.edu/FAQ/FAQformat.html#format1 . [Required]",
        required=True,
    )
    parser.add_argument(
        "-n",
        "--seqname",
        nargs="?",
        type=str,
        default=None,
        help="The prefix of fasta seq name, default is None, means to use the origin region_id of the bed",
        required=False,
    )
    parser.add_argument(
        "-f",
        "--flank",
        nargs="?",
        type=int,
        default=0,
        help="The number of the upstream and downstream slip bases, default is 0",
        required=False,
    )
    parser.add_argument(
        "-s",
        "--strand",
        type=str,
        help="Whether to fetch from strand information in bed, True/False.",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--outfile",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Output file name for storing the results, default is stdout.",
        required=False,
    )
    args = parser.parse_args()
    return args


def run(args):
    genome = Fasta(args.genome)
    bed = list(filter(lambda x: x.strip(), args.bedfile.readlines()))
    bed_list = list(map(lambda x: x.strip().split(), bed))

    result = map(
        lambda i: ">{0}({1})\n{2}".format(
            bed_list[i][3] if args.seqname is None else i + 1,
            bed_list[i][5] if args.strand == "True" else "?",
            genome.sequence(
                {
                    "chr": bed_list[i][0],
                    "start": int(bed_list[i][1]) - args.flank,
                    "stop": int(bed_list[i][2]) + args.flank,
                    "strand": bed_list[i][5] if args.strand == "True" else "+",
                }
            ).upper(),
        ),
        range(len(bed_list)),
    )

    if args.outfile:
        args.outfile.write("\n".join(result))
    else:
        print("".join(result))


if __name__ == "__main__":
    args = get_opt()
    run(args)



# %% import packages
import argparse
import sys
import os
import subprocess
import logging

import string
import random

import Bio
import Bio.SeqIO
from Bio.Align.substitution_matrices import Array
# %% Version information
VERSION_INFO = """
Author: MENG Howard

Version-06-TALE:
    2022-02-12
        TALE sequence alignment based on V6 code

Version-06:
    2022-01-05
        Test gap open penalty, especially in PAM distal side

    2021-12-07
        Add signal dist penalty;
        Support .bed format table
        Support Python3.x
        Reconstruction all code!

Version-05:
    2021-11-26
        Add signal dist penalty;
        Support .bed format table

Version-04:
    2020-09-02
        Fix output region_id 'nan' error.

Version-03:
    2020-05-22
        Add select alignment mode function

Version-02:
    2020-03-06
        Fix d=0 issue and add alignment strategy option

Version-01:
    2019-12-03
        mpmat file convert to alignment result table (.art) file

E-Mail: meng_howard@126.com

"""
# %% Learning Part START --------------------------------------------------------
LEARNING_PART = """

There will be a lot of parameters to set,

but default parameters can handle output figure in most cases.

"""

PIPELINE = """

INPUT
    <mpmat> or <BED> format

    <fasta>

    <alignment settings>

    <id_prefix>

OUTPUT
    <alignment_result_table>
        format like:


DESIGN
    0. use bedtools getfasta to get fasta temp files

    1. read all mpmat and fa file

    2. seed alignment (+/-)

    3. extend alignment (+/-)

    4. select the best alignment

    5. calculate output info (.art) file
"""
# %% function part


def _log_cmd_str(args):
    """
    INPUT:
        <args>
            obj, argparse obj

    RETURN:
        <full_cmd_str>
            str, record full command str info
    """

    full_cmd_str = """python mpmat-to-art.py
    --input {input}
    --query_seq {query_seq}
    --output {output}
    --reference {reference}
    --extend_method {extend_method}
    --extend_length {extend_length}
    --input_filetype {input_filetype}
    --input_header {input_header}
    --mpmat_fwd_mut_type {mpmat_fwd_mut_type}
    --mpmat_rev_mut_type {mpmat_rev_mut_type}
    --bedtools_path {bedtools_path}
    --align_settings {align_settings}
    --sub_mat_alphabet {sub_mat_alphabet}
    --sub_mat_specific {sub_mat_specific}
    --N0_T_award {N0_T_award}
    --temp_dir {temp_dir}
    --verbose {verbose}""".format(
        input=args.input_bam,
        query_seq=args.query_seq,
        output=args.output_bam,
        reference=args.reference,
        extend_method=args.extend_method,
        extend_length=args.extend_length,
        input_filetype=args.input_filetype,
        input_header=args.input_header,
        mpmat_fwd_mut_type=args.mpmat_fwd_mut_type,
        mpmat_rev_mut_type=args.mpmat_rev_mut_type,
        bedtools_path=args.bedtools_path,
        align_settings=args.align_settings,
        sub_mat_alphabet=args.sub_mat_alphabet,
        sub_mat_specific=args.sub_mat_specific,
        N0_T_award=args.N0_T_award,
        temp_dir=args.temp_dir,
        verbose=args.verbose
    )

    return full_cmd_str


def find_highest_mut_site(mpmat_line_list, find_method="score"):
    """
    INPUT:
        <mpmat_row>
            mpmat_row, pandas Series obj, with default column names

        <find_method>
            1. ratio
                return 5' site with highest mutation ratio;

            2. mut_count
                return 5' site with highest mutation reads count;

            3. score
                    score = mut_count * ratio ^ 2
                return 5' site with highest score;
    """
    site_num = int(mpmat_line_list[3])

    # fetch site index
    site_index_list = mpmat_line_list[6].split(",")

    # fetch SNP state
    snp_ann_list = mpmat_line_list[10].split(",")

    # fetch count and ratio
    mut_base_num_list = list(map(int, mpmat_line_list[7].split(",")))
    cover_base_num_list = list(map(int, mpmat_line_list[8].split(",")))
    mut_ratio_list = list(map(float, mpmat_line_list[9].split(",")))

    # select site index
    highest_site_index = site_index_list[0]
    highest_index = 0

    if find_method == "ratio":
        highest_index = mut_ratio_list.index(max(mut_ratio_list))

    elif find_method == "mut_count":
        highest_index = mut_base_num_list.index(max(mut_base_num_list))

    elif find_method == "score":
        mut_score_list = []
        for i in range(site_num):
            if snp_ann_list[i] != "True":
                mut_score = mut_ratio_list[i] * mut_ratio_list[i] * mut_base_num_list[i]
                mut_score_list.append(mut_score)
            else:
                mut_score_list.append(0)
        highest_index = mut_score_list.index(max(mut_score_list))

    # get site_index
    highest_site_index = site_index_list[highest_index]

    # return part
    chr_name, chr_pos = highest_site_index.split("_")[:2]
    mut_count = mut_base_num_list[highest_index]
    mut_ratio = mut_ratio_list[highest_index]
    cover_count = cover_base_num_list[highest_index]

    return chr_name, int(chr_pos), highest_site_index, mut_count, cover_count, mut_ratio


def get_mut_type_count(site_index_list_str):
    """
    INPUT
        <site_index_list_str>
            A str like 'chr1_143265617_CT,chr1_143265637_CT,chr1_143265643_CT,chr1_143265655_CT'

    RETURN
        count mut_type and return the max count mut_type
    """

    mut_type_dict = {}

    site_index_list = site_index_list_str.split(",")

    for site_index in site_index_list:
        mut_type = site_index.split("_")[-1]
        if mut_type_dict.get(mut_type) is None:
            mut_type_dict[mut_type] = 1
        else:
            mut_type_dict[mut_type] += 1

    return sorted(mut_type_dict.items(), key=lambda i: i[1], reverse=True)[0][0]


def make_temp_fasta(input_filename,
                    input_filetype,
                    genome_fasta,
                    temp_dir,
                    mut_type_rev="GA",
                    bedtools_path="bedtools",
                    header_state=False,
                    extend_region_size=20,
                    extend_region_method="highest_site",
                    log_verbose=3):
    """
    INPUT:
        <input_filename>
            str, .bed file or .mpmat file

        <input_filetype>
            str, 'bed' or 'mpmat'

        <genome_fasta>
            str, reference genome fasta file

        <temp_dir>
            str, temp dir to store temp files

        <header_state>
            bool, if input file contains a header line

        <extend_region_size>
            int,

        <select_region_method>
            1. region
            2. upstream_site
            3. downstream_site
            4. highest_site

    RETURN:
        If successful return
            (0, temp_fasta_file, temp_bed_file)

        If unsuccessful return
            (1, None, None)

    """
    # logging
    logging.basicConfig(level=(4 - log_verbose) * 10,
                        format='%(levelname)-5s @ %(asctime)s: %(message)s ',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        stream=sys.stderr,
                        filemode="w")

    # make temp BED file
    temp_bed_basename = "temp" + "_" + "".join(
        random.sample(string.ascii_letters + string.digits, 16)) + ".bed"
    in_fun_temp_bed_filename = os.path.abspath(os.path.join(temp_dir, temp_bed_basename))
    temp_bed_file = open(in_fun_temp_bed_filename, "w")

    with open(input_filename, "r") as in_file:
        if header_state:
            line = in_file.readline()

        for line in in_file:
            line_list = line.split("\t")

            region_chr = line_list[0]
            region_start = int(line_list[1])
            region_end = int(line_list[2])
            region_index = "%s_%s_%s" % (region_chr, region_start, region_end)

            # get region strand
            strand_info = "+"
            if input_filetype == "bed":
                if len(line_list) >= 6:
                    if line_list[5] == "-":
                        strand_info = "-"

            elif input_filetype == "mpmat":
                region_mut_type = get_mut_type_count(line_list[6])
                if region_mut_type == mut_type_rev:
                    strand_info = "-"

            # extend region
            if extend_region_method == "region":
                region_extend_start = region_start - extend_region_size
                region_extend_end = region_end + extend_region_size

            elif extend_region_method == "upstream_site":
                if strand_info == "+":
                    region_extend_start = region_start - extend_region_size
                    region_extend_end = region_start + extend_region_size
                else:
                    region_extend_start = region_end - extend_region_size
                    region_extend_end = region_end + extend_region_size

            elif extend_region_method == "downstream_site":
                if strand_info == "+":
                    region_extend_start = region_end - extend_region_size
                    region_extend_end = region_end + extend_region_size
                else:
                    region_extend_start = region_start - extend_region_size
                    region_extend_end = region_start + extend_region_size

            elif extend_region_method == "highest_site":
                if input_filetype == "mpmat":
                    site_chr_name, site_chr_pos = find_highest_mut_site(mpmat_line_list=line_list, find_method="score")
                    region_extend_start = site_chr_pos - extend_region_size
                    region_extend_end = site_chr_pos + extend_region_size

                else:
                    region_extend_start = region_start - extend_region_size
                    region_extend_end = region_end + extend_region_size

            else:
                raise KeyError("Wrong extend method!")

            # output info into bed
            bed_out_list = [region_chr, region_extend_start, region_extend_end, region_index, 1, strand_info]
            temp_bed_file.write("\t".join(map(str, bed_out_list)) + "\n")

    temp_bed_file.close()

    # log
    logging.info("Output temp bed with filename: %s\n" % in_fun_temp_bed_filename)

    # use bedtools to make fasta file
    temp_fasta_filename = in_fun_temp_bed_filename + ".fa"

    # only report sequence same to the genome fwd
    tool_bedtools_cmd_fmt = "{bedtools} getfasta -fi {ref_fasta}  -fo {out_fasta} -bed {input_bed} -name+ "
    tool_bedtools_cmd = tool_bedtools_cmd_fmt.format(
        bedtools=bedtools_path,
        ref_fasta=genome_fasta,
        out_fasta=temp_fasta_filename,
        input_bed=in_fun_temp_bed_filename
    )

    # bedtools getfasta is 0-based index system, while UCSC is 1-based index system
    # the real index
    # chr_start = bedtools.start + 1
    # chr_end = bedtools.end

    # log
    logging.info("bedtools command:\n%s\n" % tool_bedtools_cmd)

    # run bedtools
    get_fasta_runcode = subprocess.call(tool_bedtools_cmd, shell=True)

    # return
    if get_fasta_runcode != 0:
        os.remove(temp_fasta_filename)
        os.remove(in_fun_temp_bed_filename)
        logging.error("Error! Unsuccessfully get region FASTA seq!")
        return 1, None, None

    else:
        logging.info("Successfully get region FASTA seq!\n")
        return 0, temp_fasta_filename, in_fun_temp_bed_filename


# ---------------------------------------------------------------------->>>>>>
# alignment related functions
# ---------------------------------------------------------------------->>>>>>
def _make_np_substitution_mat(mat_match_score=5, mat_mismatch_score=-4, alphabet="AGCTN", add_sub_info=""):
    """
    INPUT:
        <mat_match_score>
            int, default = 5

        <mat_mismatch_score>
            int, default = -4

        <alphabet>
            str, default = "AGCTN", use to build up score matrix

        <add_info>
            str, format like ref:query=score, A:G=5,G:A=3..., comma as sperator

    RETURN:
        <substitution_matrix>
            dict, alignment matrix for aligner
    """
    # raw matrix
    score_matrix = Array(alphabet=alphabet, dims=2, dtype="float")

    for i1, letter1 in enumerate(alphabet):
        for i2, letter2 in enumerate(alphabet):
            if letter1 == letter2:
                score_matrix[i1, i2] = mat_match_score

            elif letter1 != letter2:
                score_matrix[i1, i2] = mat_mismatch_score

    # fix align score
    if add_sub_info != "":
        try:
            add_sub_info_list = add_sub_info.split(",")
            for sub_info in add_sub_info_list:
                mut_info, align_score = sub_info.split("=")
                from_base, to_base = mut_info.split(":")
                score_matrix[from_base.upper(), to_base.upper()] = float(align_score)
        except:
            Warning("Something wrong with add_info, " + "\n")
            raise IOError("Please set add info as format ref:query=score, A:G=5,G:A=3..." + "\n")

    return score_matrix


def degen_align_obj_analysis(alignment, N0_T_award_score=5, degeneracy_base="AG"):
    """
    INPUT:
        <alignment>
            obj, alignment result

    RETURN:
        <info_dict>
            int, gap count
            int, match count
            int, mismatch count
            list, gap position [0 based index]
            list, mismatch position [0 based index]
            float, alignment score, higher means better alignment
            obj, alignment obj

    INFO:
        The gap count should be the num of gap contain in qeury alignment region.

        e.g.

        AGTGGTAGACATT
        ------||X|--|
        ------AGCC--T


        return = {
            "gap_count" : 2,
            "match_count" : 4,
            "mismatch_count" : 1,
            "gap_pos" : [10, 11],
            "mismatch_pos" : [8],
            "alignment_score" : [N0_T_award_score] + [raw alignment score]
            "alignment" : alignment obj,
            "meta": {
                "N0_T_award_score" : N0_T_award_score,
                "degeneracy_base" : degeneracy_base,
            }
        }
    """

    # init dict
    info_dict = {
        "gap_count": 0,
        "match_count": 0,
        "mismatch_count": 0,
        "degen_match_count": 0,
        "degen_count": 0,
        "degen_mismatch_count": 0,
        "gap_pos": [],
        "mismatch_pos": [],
        "degen_match_pos": [],
        "degen_mismatch_pos": [],
        "alignment_score": alignment.score,
        "alignment": alignment,
        "meta": {
            "N0_T_award_score": N0_T_award_score,
            "degeneracy_base": degeneracy_base
        }
    }

    # fetch alignment info
    alignment_list = str(alignment).split("\n")
    query_length = len(alignment.query)

    # load alignment
    ref_list = alignment_list[0]
    info_list = alignment_list[1]
    query_list = alignment_list[2]

    fix_info_list = list(info_list)

    # count init value
    count_state = False
    count_query_base = 0

    query_index = 0

    # counting
    for run_index, info_base in enumerate(info_list):
        if (not count_state) and (query_list[run_index] != "-"):
            count_state = True

            # record N0 info
            if info_dict.get("N0_base") is None:
                info_dict["N0_base"] = alignment_list[0][run_index]

                # make N0 T award
                if info_dict["N0_base"] == "T":
                    info_dict["alignment_score"] += N0_T_award_score

        if count_state:
            if info_base == "|":
                info_dict["match_count"] += 1
                info_dict["degen_match_count"] += 1
                query_index += 1

            elif info_base == "X" or info_base == ".":
                info_dict["mismatch_count"] += 1
                info_dict["mismatch_pos"].append(query_index)
                query_index += 1

                if (ref_list[run_index] == degeneracy_base[0]) and (query_list[run_index] == degeneracy_base[1]):
                    info_dict["degen_count"] += 1
                    info_dict["degen_match_count"] += 1
                    info_dict["degen_match_pos"].append(query_index)
                    fix_info_list[run_index] = "."

                else:
                    info_dict["degen_mismatch_count"] += 1
                    info_dict["degen_mismatch_pos"].append(query_index)
                    fix_info_list[run_index] = "X"

            elif info_base == "-":
                info_dict["gap_count"] += 1
                info_dict["gap_pos"].append(query_index)

                if query_list[run_index] != "-":
                    query_index += 1

        if query_index >= query_length:
            break

    # add dict
    info_dict["alignment_degen_path"] = "".join(fix_info_list)

    if query_index < query_length:
        info_dict["alignment_score"] -= 20

    return info_dict


def run_step_align(align_ref_seq, align_query_seq, aligner,
                   align_window_size=25,
                   align_step_size=13,
                   N0_T_award_score=5,
                   degeneracy_base="AG"):
    """
    INPUT:
        <align_ref_seq>
            str, reference sequence

        <align_query_seq>
            str, align sequence

        <aligner>
            obj, Bio PairwiseAligner with setting score matrix

    RETURN
        <align_res_list>
            list, each item is also a list like:
                [
                    str.full_ref_seq,
                    str.full_path,
                    str.full_query_seq,
                    int.rel_align_start_idx,
                    int.rel_align_end_idx,
                    int.align_res.score,
                    int.start_idx,
                    int.end_idx,
                    obj.align_res
                ]

    """
    # convert Seq obj
    align_ref_seq = Bio.Seq.Seq(align_ref_seq)
    align_query_seq = Bio.Seq.Seq(align_query_seq)

    # return list
    align_res_list = []

    # make align steps
    region_idx_list = []
    start_idx = 0
    while start_idx < len(align_ref_seq):
        end_idx = start_idx + align_window_size

        if end_idx <= len(align_ref_seq):
            region_idx_list.append((start_idx, end_idx))
        else:
            region_idx_list.append((start_idx, len(align_ref_seq)))
            break

        start_idx += align_step_size

    # step alignment
    for start_idx, end_idx in region_idx_list:
        region_ref_seq = align_ref_seq[start_idx: end_idx]

        # select the best hit in one alignment
        align_res_dict_list = []
        # DEBUG temp 2022-03-24
        if region_ref_seq.count('N') >= 5:
            continue
        # print(region_ref_seq, align_query_seq)
        # DEBUG temp 2022-03-24
        for step_align_res in aligner.align(region_ref_seq, align_query_seq):
            # this params is designed for SpCas9 seed region alignment
            align_res_dict = degen_align_obj_analysis(step_align_res,
                                                      N0_T_award_score=N0_T_award_score,
                                                      degeneracy_base=degeneracy_base)

            align_res_dict_list.append(align_res_dict)

        # sort
        align_res_dict_list = sorted(align_res_dict_list,
                                     key=lambda i: (i["alignment_score"],
                                                    i["degen_match_count"],
                                                    i["match_count"]),
                                     reverse=True)

        step_best_align_dict = align_res_dict_list[0]

        step_align_res = align_res_dict_list[0]["alignment"]
        align_info_list = str(step_align_res).split("\n")
        align_count = 0

        # fix align path
        fix_align_info_path = align_res_dict_list[0]["alignment_degen_path"]

        path_start_idx = 0
        path_end_idx = len(align_info_list[0])

        path_start_state = False
        path_end_state = False

        for path_idx, path_info in enumerate(align_info_list[1]):
            # query_info = align_info_list[2][path_idx]
            query_info = fix_align_info_path[path_idx]

            if (not path_start_state) and (query_info != "-"):
                path_start_idx = path_idx
                path_start_state = True

            if path_start_state:
                if align_info_list[2][path_idx] != "-":
                    align_count += 1

            if align_count >= len(align_query_seq):
                if path_start_state and (not path_end_state) and query_info == "-":
                    path_end_idx = path_idx
                    path_end_state = True

            if path_start_state and path_end_state:
                break

        # make full length alignment result
        full_ref_seq = align_ref_seq[:start_idx] + "".join(align_info_list[0]) + align_ref_seq[end_idx - 1:]
        # full_path = "-" * start_idx + "".join(align_info_list[1]) + "-" * (len(align_ref_seq) - end_idx)
        full_path = "-" * start_idx + "".join(fix_align_info_path) + "-" * (len(align_ref_seq) - end_idx)
        full_query_seq = "-" * start_idx + "".join(align_info_list[2]) + "-" * (len(align_ref_seq) - end_idx)

        # make relative index
        rel_align_start_idx = path_start_idx + start_idx
        rel_align_end_idx = path_end_idx + start_idx

        # make dict
        step_best_align_dict["align_info_ref"] = str(full_ref_seq)
        step_best_align_dict["align_info_path"] = str(full_path)
        step_best_align_dict["align_info_query"] = str(full_query_seq)
        step_best_align_dict["align_info_start_idx"] = rel_align_start_idx
        step_best_align_dict["align_info_end_idx"] = rel_align_end_idx
        step_best_align_dict["step_start_idx"] = start_idx
        step_best_align_dict["step_end_idx"] = end_idx

        # back list
        align_res_list.append(step_best_align_dict)

    return align_res_list


def get_dist_to_signal(
        signal_chr_start,
        signal_chr_end,
        signal_strand,
        in_align_chr_start,
        in_align_chr_end,
        align_reverse=False):
    """
    INPUT
        <signal_chr_start>
            int,
        <signal_chr_end>
            int,
        <align_chr_start>
            int,
        <align_chr_end>
            int,
        <align_reverse>
            bool, True means aligned to the - strand

    RETURN:
        the distance between aligned TALE 3' site to signal 5' site

    INFO:

        Example1:

            REF - - - - - - - - - - - - 5'C C C C C C C 3' -  -  -  -
            TALE  5'-CGATGGCTATTT-3'
                                  1 2 3   4
            RETURN dist=4

        Example2:

            REF - - - - - - - - - - - - 3'G G G G G G G 5' -  -  -  -
            TALE  5'- T T G T T T T-3'
                               -1 0 1 2 3 4 5 6 7 8 9 10
            RETURN dist=10
    """

    if signal_strand == "+":
        if not align_reverse:
            return signal_chr_start - in_align_chr_end

        else:
            return in_align_chr_start - signal_chr_start

    else:
        if not align_reverse:
            return signal_chr_end - in_align_chr_end

        else:
            return in_align_chr_start - signal_chr_end


def get_align_chrom_coordinate(align_res_dict,
                               ref_chr_start,
                               ref_chr_end,
                               align_ref_start,
                               align_ref_end,
                               reverse_state=False):
    """
    INPUT:
        <align_res_dict>
            dict, format like

            {
                'N0_base': 'T',
                 'align_info_end_idx': 83,
                 'align_info_path': '-----------------------------------------------------------------------.||||.||||||-------------------------------------------------------',
                 'align_info_query': '-----------------------------------------------------------------------CGATGGCTATTT-------------------------------------------------------',
                 'align_info_ref': 'AGTTCAAACGAGGATTCAGGTTTTAGGCCAAATAAGAGATAACATTCTTGAATTCATAATTGCTAGAATCATGATGACTATTTGAAAGGATATATAATTTATCCCTTAATAACTGAAAGGTGTCATCCAAAACTGTTGT',
                 'align_info_start_idx': 71,
                 'alignment': <Bio.Align.PairwiseAlignment object at 0x7fb451d74e20>,
                 'alignment_score': 54.0,
                 'degen_count': 1,
                 'degen_match_count': 11,
                 'degen_match_pos': [5],
                 'degen_mismatch_count': 1,
                 'degen_mismatch_pos': [0],
                 'gap_count': 0,
                 'gap_pos': [],
                 'match_count': 10,
                 'meta': {'N0_T_award_score': 5, 'degeneracy_base': 'AG'},
                 'mismatch_count': 2,
                 'mismatch_pos': [0, 5],
                 'step_end_idx': 85,
                 'step_start_idx': 60
            }

        <ref_chr_start>
            int, UCSC chromosome index of ref sequence start

        <ref_chr_end>
            int, UCSC chromosome index of ref sequence end

        <align_ref_start>
            int, align index on ref seq 0-based

        <align_ref_end>
            int, align index on ref seq 0-based

        <reverse_state>
            bool, if align to the reverse strand

    """

    total_gap_count = align_res_dict["align_info_ref"][align_ref_start:align_ref_end].count("-")

    if not reverse_state:
        back_align_chr_start = ref_chr_start + align_ref_start
        back_align_chr_end = ref_chr_start + align_ref_end - total_gap_count - 1

    else:
        back_align_chr_end = ref_chr_end - align_ref_start
        back_align_chr_start = ref_chr_end - align_ref_end + 1 + total_gap_count

    return back_align_chr_start, back_align_chr_end


def report_TALE_align_res(TALE_align_dict):
    """
    INPUT:
        <TALE_align_dict>
            dict,

            {
                'N0_base': 'T',
                 'align_info_end_idx': 83,
                 'align_info_path': '-----------------------------------------------------------------------.||||.||||||-------------------------------------------------------',
                 'align_info_query': '-----------------------------------------------------------------------CGATGGCTATTT-------------------------------------------------------',
                 'align_info_ref': 'AGTTCAAACGAGGATTCAGGTTTTAGGCCAAATAAGAGATAACATTCTTGAATTCATAATTGCTAGAATCATGATGACTATTTGAAAGGATATATAATTTATCCCTTAATAACTGAAAGGTGTCATCCAAAACTGTTGT',
                 'align_info_start_idx': 71,
                 'alignment': <Bio.Align.PairwiseAlignment object at 0x7fb451d74e20>,
                 'alignment_score': 54.0,
                 'degen_count': 1,
                 'degen_match_count': 11,
                 'degen_match_pos': [5],
                 'degen_mismatch_count': 1,
                 'degen_mismatch_pos': [0],
                 'gap_count': 0,
                 'gap_pos': [],
                 'match_count': 10,
                 'meta': {'N0_T_award_score': 5, 'degeneracy_base': 'AG'},
                 'mismatch_count': 2,
                 'mismatch_pos': [0, 5],
                 'step_end_idx': 85,
                 'step_start_idx': 60
            }

    RETURN:
        <align_out_list>
            list, .art format list
            "chrom",
            "start",
            "end",
            "region_index",
            "align_chr_name",
            "align_chr_start",
            "align_chr_end",
            "align_strand",
            "align_dist_to_signal",
            "align_N0_base",
            "align_total_match",
            "align_total_mismatch",
            "align_degen_total_match",
            "align_degen_total_mismatch",
            "align_degen_num",
            "align_total_gap",
            "align_score",
            "align_target_seq",
            "align_info_state",
            "align_query_seq"
    """

    # load index
    start_idx = TALE_align_dict["align_info_start_idx"]
    end_idx = TALE_align_dict["align_info_end_idx"]

    # make list
    return_out_list = [
        TALE_align_dict["region_info"][0],
        TALE_align_dict["region_info"][1],
        TALE_align_dict["region_info"][2],
        TALE_align_dict["region_info"][3],
        TALE_align_dict["align_chr_name"],
        TALE_align_dict["align_chr_start"],
        TALE_align_dict["align_chr_end"],
        TALE_align_dict["align_chr_strand"],
        TALE_align_dict["align_dist_to_signal"],
        TALE_align_dict["N0_base"],
        TALE_align_dict["match_count"],
        TALE_align_dict["mismatch_count"],
        TALE_align_dict["degen_match_count"],
        TALE_align_dict["degen_mismatch_count"],
        TALE_align_dict["degen_count"],
        TALE_align_dict["gap_count"],
        TALE_align_dict["alignment_score"],
        TALE_align_dict["align_info_ref"][start_idx:end_idx],
        TALE_align_dict["align_info_path"][start_idx:end_idx],
        TALE_align_dict["align_info_query"][start_idx:end_idx]
    ]

    return return_out_list


# %% main
if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="From .mpmat to alignment result table (.art) file")

    parser.add_argument("-i", "--input",
                        help=".mpmat table or .bed table, generated from <pmat-merge.py> or <mpmat-select.py>",
                        required=True)

    parser.add_argument("-q", "--query_seq",
                        help="TALE_seq with N0 base info", required=True)

    parser.add_argument("-o", "--output",
                        help="Output alignment result table (.art) filename, default=stdout", default="stdout")

    parser.add_argument("-r", "--reference",
                        help="Reference genome fasta file", required=True)

    parser.add_argument("-m", "--extend_method",
                        help="Select and extend region of mpmat file to get FASTA file, "
                             "<region> <upstream_site> <downstream_site> <highest_site>, "
                             "default=region", default="region")

    parser.add_argument("-e", "--extend_length",
                        help="Region extend length, default=40", default=40, type=int)

    parser.add_argument("--input_filetype",
                        help="'mpmat' or 'bed', default='mpmat'", default="mpmat")

    parser.add_argument("--input_header",
                        help="If .mpmat file contain header, default=False", default="False")

    parser.add_argument("--mpmat_fwd_mut_type",
                        help="mpmat file fwd strand mutation type, default=CT", default="CT")

    parser.add_argument("--mpmat_rev_mut_type",
                        help="mpmat file fwd strand mutation type, default=GA", default="GA")

    parser.add_argument("--bedtools_path",
                        help="Software <bedtools> PATH, default considered <bedtools> is already included in current "
                             "PATH environment.",
                        default="bedtools")

    parser.add_argument("--align_settings",
                        help="Set <align_match_score> <align_mismatch_score> <align_gap_open_score> "
                             "<align_gap_extension_score>, default=5,-4,-24,-8",
                        default="5,-4,-24,-8")

    parser.add_argument("--sub_mat_specific",
                        help="Set specific alignment score, format like ref:query=score, default='A:G=3'",
                        default="A:G=3")

    parser.add_argument("--sub_mat_alphabet",
                        help="substitution matrix alphabet, default=ATCG",
                        default="AGCT")

    parser.add_argument("--N0_T_award",
                        help="TALE prefer 'T' at N0 position, there will be a award score if a 'T' at N0, default=10",
                        default=10, type=int)

    parser.add_argument("--temp_dir",
                        help="default is the same dir to --input", default=None)

    parser.add_argument("--verbose",
                        help="A larger value means more details, default=3", default=3, type=int)

    ARGS = parser.parse_args()

    # ------------------------------------------------------------------------>>>>>>>
    # logging
    # ------------------------------------------------------------------------>>>>>>>
    logging.basicConfig(level=(4 - ARGS.verbose) * 10,
                        format='%(levelname)-5s @ %(asctime)s: %(message)s ',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        stream=sys.stderr,
                        filemode="w")

    # ------------------------------------------------------------------------>>>>>>>
    # get temp dir check output file dir
    # ------------------------------------------------------------------------>>>>>>>
    if ARGS.output == "stdout":
        temp_out_dir = os.getcwd()
    else:
        temp_out_dir = os.path.dirname(ARGS.output)

    ARGS.temp_dir = temp_out_dir

    if ARGS.input_header == "True":
        ARGS.input_header = True
    else:
        ARGS.input_header = False

    # ------------------------------------------------------------------------>>>>>>>
    # output full cmd
    # ------------------------------------------------------------------------>>>>>>>
    sys.stderr.write("-" * 80 + "\n")
    sys.stderr.write(_log_cmd_str(ARGS) + "\n")

    # ------------------------------------------------------------------------>>>>>>>
    # make temp fasta
    # ------------------------------------------------------------------------>>>>>>>
    # get fasta temp
    sys.stderr.write("-" * 80 + "\n")
    logging.info("From input file to generate FASTA file...\n")
    temp_fa_state, temp_fa_filename, temp_bed_filename = make_temp_fasta(ARGS.input,
                                                                         ARGS.input_filetype,
                                                                         ARGS.reference,
                                                                         temp_out_dir,
                                                                         mut_type_rev=ARGS.mpmat_rev_mut_type,
                                                                         bedtools_path=ARGS.bedtools_path,
                                                                         header_state=ARGS.input_header,
                                                                         extend_region_size=ARGS.extend_length,
                                                                         extend_region_method=ARGS.extend_method,
                                                                         log_verbose=ARGS.verbose)

    if temp_fa_state != 0:
        logging.error("Make FASTA file unsuccessfully!")
        raise IOError()
    else:
        logging.info("Make FASTA file successfully!")

    # ------------------------------------------------------------------------>>>>>>>
    # make substitution matrix
    # ------------------------------------------------------------------------>>>>>>>
    align_match, align_mismatch, align_gap_open, align_gap_extension = map(int, str(ARGS.align_settings).split(","))

    align_score_matrix = _make_np_substitution_mat(mat_match_score=align_match,
                                                   mat_mismatch_score=align_mismatch,
                                                   alphabet=ARGS.sub_mat_alphabet,
                                                   add_sub_info=ARGS.sub_mat_specific)

    sys.stderr.write("-" * 80 + "\n")
    sys.stderr.write("Aligner substitution matrix:\n")
    sys.stderr.write(str(align_score_matrix))

    # ------------------------------------------------------------------------>>>>>>>
    # make aligner
    # ------------------------------------------------------------------------>>>>>>>
    # reference Table 6.1: Meta-attributes of the pairwise aligner objects.
    # http://biopython.org/DIST/docs/tutorial/Tutorial.html#sec118
    TALE_aligner = Bio.Align.PairwiseAligner()
    TALE_aligner.substitution_matrix = align_score_matrix
    TALE_aligner.open_gap_score = align_gap_open
    TALE_aligner.extend_gap_score = align_gap_extension
    TALE_aligner.query_left_gap_score = 0
    TALE_aligner.query_right_gap_score = 0
    TALE_aligner.mode = "global"

    sys.stderr.write("-" * 80 + "\n")
    sys.stderr.write("TALE aligner:" + "\n")
    sys.stderr.write(str(TALE_aligner))
    sys.stderr.write("-" * 80 + "\n")

    # ------------------------------------------------------------------------>>>>>>>
    # output header
    # ------------------------------------------------------------------------>>>>>>>
    out_header_list = [
        "chrom",
        "start",
        "end",
        "region_index",
        "align_chr_name",
        "align_chr_start",
        "align_chr_end",
        "align_strand",
        "align_dist_to_signal",
        "align_N0_base",
        "align_total_match",
        "align_total_mismatch",
        "align_degen_total_match",
        "align_degen_total_mismatch",
        "align_degen_num",
        "align_total_gap",
        "align_score",
        "align_target_seq",
        "align_info_state",
        "align_query_seq"
    ]

    # ------------------------------------------------------------------------>>>>>>>
    # open files
    # ------------------------------------------------------------------------>>>>>>>
    if ARGS.output == "stdout":
        out_file = sys.stdout
    else:
        out_file = open(ARGS.output, "w")

    region_file = open(ARGS.input, "r")
    temp_bed = open(temp_bed_filename, "r")
    temp_fa = Bio.SeqIO.parse(temp_fa_filename, format="fasta")

    # output header line
    out_file.write("\t".join(out_header_list) + "\n")

    # ------------------------------------------------------------------------>>>>>>>
    # make query sequence
    # ------------------------------------------------------------------------>>>>>>>
    query_seq = Bio.Seq.Seq(str(ARGS.query_seq).upper())

    # ------------------------------------------------------------------------>>>>>>>
    # alignment
    # ------------------------------------------------------------------------>>>>>>>
    if ARGS.input_header:
        region_file.readline()

    for fa_idx, fa_info in enumerate(temp_fa):
        bed_list = temp_bed.readline().strip().split("\t")
        region_list = region_file.readline().strip().split("\t")

        # load genome coordinate of bedtools getfasta
        bed_chr_start = int(bed_list[1]) + 1
        bed_chr_end = int(bed_list[2])

        # load signal region coordinate
        region_chr_start = int(region_list[1])
        region_chr_end = int(region_list[2])
        region_signal_strand = bed_list[5]

        # load fasta seq
        fasta_seq_fwd = fa_info.seq.upper()
        fasta_seq_rev = Bio.Seq.reverse_complement(fasta_seq_fwd)

        # step alignment
        step_window_size = 50
        step_move_size = 10

        step_align_res_fwd = run_step_align(align_ref_seq=fasta_seq_fwd,
                                            align_query_seq=query_seq,
                                            aligner=TALE_aligner,
                                            align_window_size=step_window_size,
                                            align_step_size=step_move_size,
                                            N0_T_award_score=ARGS.N0_T_award,
                                            degeneracy_base="AG")

        step_align_res_rev = run_step_align(align_ref_seq=fasta_seq_rev,
                                            align_query_seq=query_seq,
                                            aligner=TALE_aligner,
                                            align_window_size=step_window_size,
                                            align_step_size=step_move_size,
                                            N0_T_award_score=ARGS.N0_T_award,
                                            degeneracy_base="AG")

        # select the best for each strand
        # DEBUG temp 2022-03-24
        if not step_align_res_fwd:
            continue
        # DEBUG temp 2022-03-24
        best_align_dict_fwd = sorted(step_align_res_fwd,
                                     key=lambda i: (i["alignment_score"],
                                                    i["degen_match_count"],
                                                    i["match_count"]),
                                     reverse=True)[0]

        best_align_dict_rev = sorted(step_align_res_rev,
                                     key=lambda i: (i["alignment_score"],
                                                    i["degen_match_count"],
                                                    i["match_count"]),
                                     reverse=True)[0]

        # select the best
        best_align_dict = None
        best_align_reverse_state = None

        if best_align_dict_fwd["alignment_score"] > best_align_dict_rev["alignment_score"]:
            best_align_dict = best_align_dict_fwd
            best_align_reverse_state = False

        else:
            best_align_dict = best_align_dict_rev
            best_align_reverse_state = True

        # get genome coordinate
        align_chr_start, align_chr_end = get_align_chrom_coordinate(best_align_dict,
                                                                    ref_chr_start=bed_chr_start,
                                                                    ref_chr_end=bed_chr_end,
                                                                    align_ref_start=best_align_dict[
                                                                        "align_info_start_idx"],
                                                                    align_ref_end=best_align_dict["align_info_end_idx"],
                                                                    reverse_state=best_align_reverse_state)

        # calculate signal to dist penalty
        align_dist_to_signal = get_dist_to_signal(signal_chr_start=region_chr_start,
                                                  signal_chr_end=region_chr_end,
                                                  signal_strand=region_signal_strand,
                                                  in_align_chr_start=align_chr_start,
                                                  in_align_chr_end=align_chr_end,
                                                  align_reverse=best_align_reverse_state)

        # add info into dict
        best_align_dict["align_chr_name"] = bed_list[0]
        best_align_dict["align_chr_start"] = align_chr_start
        best_align_dict["align_chr_end"] = align_chr_end
        best_align_dict["align_dist_to_signal"] = align_dist_to_signal
        best_align_dict["region_info"] = bed_list[:4]

        if not best_align_reverse_state:
            best_align_dict["align_chr_strand"] = "+"
        else:
            best_align_dict["align_chr_strand"] = "-"

        out_list = report_TALE_align_res(best_align_dict)

        # output
        out_file.write("\t".join(list(map(str, out_list))) + "\n")

    # ------------------------------------------------------------------------>>>>>>>
    # close and  remove temp files
    # ------------------------------------------------------------------------>>>>>>>
    if ARGS.output != "stdout":
        out_file.close()

    temp_bed.close()
    region_file.close()

    # remove temp file
    os.remove(temp_bed_filename)
    os.remove(temp_fa_filename)

    # ------------------------------------------------------------------------>>>>>>>
    # final log
    # ------------------------------------------------------------------------>>>>>>>
    logging.info("Done!")

# 2022-02-13 by Meng Haowei



#!/usr/bin/env python
# encoding: utf-8
import argparse

import pandas as pd


def get_parser():
    """Get param and return a parser object."""
    parser = argparse.ArgumentParser(
        description="convert MGI md5 file to md5sum supported file"
    )
    parser.add_argument("filename", type=str, help="input MGI md5 file")
    return parser


def main(filename):
    """Read mgi-like md5 file and convert to a normal md5 file."""
    df = pd.read_csv(
        filename,
        header=None,
        index_col=False,
        sep="  ",
        engine="python",
    )
    df.columns = ["md5", "filename"]
    df.md5 = df.md5.map(lambda x: x.lower())
    df.filename = df.filename + ".gz"
    df.to_csv(
        filename.replace(".txt", "") + ".fix.md5", header=False, index=False, sep="\t"
    )


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(filename=args.filename)



import logging
import sys

import pysam

u"""说明文档
本文件中构造了用于解析pysam.AlignedSegment对象中
"""


def back_indel_shift(info_index_list, cur_index) -> int:
    """return acc_shift(back_indel_shift)

    Args:
        info_index_list (list/tuple): list or tuples generated from align.cigar tuples
        cur_index (int): index related to MD tag in BAM file

    Returns:
        int: acc_shift index
    """
    # parse soft clip and insertion
    if len(info_index_list) == 0:
        return 0

    acc_shift = 0
    for info_start_index, info_len in info_index_list:
        if info_start_index >= cur_index:
            return acc_shift
        else:
            acc_shift += info_len
    return acc_shift


def get_align_mismatch_pairs(align, ref_genome_dict=None) -> list:
    """input a pysam AlignedSegment object

    Args:
        align (pysam.AlignedSeqment object): pysam.AlignedSeqment object
        ref_genome_dict (dict, optional): returned dict from load_reference_fasta_as_dict(). Defaults to None.

    Returns:
        list/None:
            it returns mismatch_pair_list, just like [ref_index, align_index, ref_base, align_base];
            and the "ref_index" is the same coordinate with UCSC genome browser;
            When NM == 0, it returns None.
    """
    # No mismatch
    try:
        if align.get_tag("NM") == 0:
            return None
    except:
        return None

    MD_tag_state = align.has_tag("MD")

    if MD_tag_state:
        # parse softclip, insertion and deletion
        info_index_list = []
        accu_index = 0

        for cigar_type, cigar_len in align.cigartuples:
            if cigar_type == 1 or cigar_type == 4:
                info_index_list.append((accu_index + 1, cigar_len))

            elif cigar_type == 2:
                info_index_list.append((accu_index + 1, -cigar_len))

            accu_index += cigar_len

        # parse MD tag
        mismatch_pair_list = []
        cur_base = ""
        cur_index = 0
        bases = align.get_tag("MD")

        i = 0
        while i < len(bases):
            base = bases[i]

            if base.isdigit():
                cur_base += base
                i += 1

            else:
                cur_index += int(cur_base)
                cur_base = ""

                if base == "^":
                    i += 1
                    del_str = ""

                    while (bases[i].isalpha()) and (i < len(bases)):
                        del_str += bases[i]
                        i += 1

                    cur_index += len(del_str)
                    del_str = ""

                elif base.isalpha():
                    cur_index += 1
                    ref_base = base
                    i += 1

                    # add into list
                    fix_index = cur_index + back_indel_shift(info_index_list, cur_index)

                    if fix_index < len(align.query_sequence):
                        mismatch_pair_list.append(
                            [
                                cur_index + align.reference_start,
                                cur_index - 1,
                                ref_base,
                                align.query_sequence[fix_index - 1],
                            ]
                        )
                    else:
                        return None

        return mismatch_pair_list
    else:
        mismatch_pair_list = []
        for align_idx, ref_idx in align.get_aligned_pairs():
            if (align_idx is not None) and (ref_idx is not None):
                align_base = align.query_sequence[align_idx]
                ref_base = ref_genome_dict[align.reference_name][ref_idx]

                if align_base != ref_base:
                    mismatch_pair_list.append(
                        [ref_idx + 1, align_idx, ref_base, align_base]
                    )

        return mismatch_pair_list


def get_query_site_mpileup_info(
    site_idx_list,
    bam_obj,
    genome_obj,
    ignore_overlaps=False,
    min_base_quality=20,
    min_mapping_quality=50,
    mpileup_extend_length=10,
    dist_cutoff=200,
    max_depth=4000,
    log_verbose=logging.DEBUG,
) -> dict:
    """get each base mutation count info from the input list (site_idx_list)
    Args:
        site_idx_list (list):
            It is a list format like:
                ["chr1_125178576_CT", "chr1_125178578_CT", "chr1_125178580_CT", "chr1_125178588_CA"]
                or
                ["chr1_125178576", "chr1_125178578", "chr1_125178580", "chr1_125178588"]
            The site coordinate is related to the UCSC genoem browser index.
            Site index have to be sorted as ascending order.
        bam_obj (pysam.AlignmentFile object):
            pysam.AlignmentFile() object
        genome_obj (pysam.FastaFile object):
            genome object, which is created by pysam.FastaFile()
        ignore_overlaps (bool, optional):
            Whether to ignore overlapping regions.
            Defaults to False.
        min_base_quality (int, optional):
            Minimum base quality.
            Defaults to 20.
        min_mapping_quality (int, optional):
            Minimum mapping quality.
            Defaults to 50.
        mpileup_extend_length (int, optional):
            To prevent reads from being missed, set an appropriate number to extend fetching region.
            Defaut number is fun.
            Defaults to 10.
        dist_cutoff (int, optional):
            The number of bases to be processed at one time.
            An appropriate number, can improve performance.
            Defaults to 200.
        max_depth (int, optional):
            Pileup file max depth.
            Defaults to 4000.
        log_verbose (logging object, optional):
            A logging states object, which can be logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL.
            Defaults to logging.DEBUG.
    Raises:
        IOError: [description]
        IOError: [description]

    Returns:
        dict/None:
            key:
                site_index only key chr_name and chr_pos like "chr1_125178576" rather than "chr1_125178576_CT"
            value:
                A, T, C, G, N, other count
            If an empty list is given, it will return None.
    """
    # log setting
    logging.basicConfig(
        level=log_verbose,
        format="%(levelname)-5s @ %(asctime)s: %(message)s ",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        filemode="w",
        force=True,
    )
    # check site index list
    if len(site_idx_list) == 0:
        return None

    chr_name = site_idx_list[0].split("_")[0]
    region_start = region_end = int(site_idx_list[0].split("_")[1])
    site_dist = 0

    for site_idx in site_idx_list:
        if site_idx.split("_")[0] != chr_name:
            logging.error("<site_idx_list> error: %s!" % site_idx)
            raise IOError(
                "<site_idx_list> all site index have to on the same chromosome."
            )

    if len(site_idx_list) >= 2:
        region_start = int(site_idx_list[0].split("_")[1])
        region_end = int(site_idx_list[-1].split("_")[1])
        site_dist = region_end - region_start

    if site_dist >= dist_cutoff:
        raise IOError(
            "<site_idx_list> mpileup region is too large, which could take a lot of time!"
        )

    # make raw dict
    site_dict = {}
    for site_idx in site_idx_list:
        key = "_".join(site_idx.split("_")[:2])
        site_dict[key] = {"A": 0, "T": 0, "C": 0, "G": 0, "N": 0, "total": 0}

    # run mpileup
    mpileup_iter = bam_obj.pileup(
        chr_name,
        region_start - mpileup_extend_length,
        region_end + mpileup_extend_length,
        fastafile=genome_obj,
        max_depth=max_depth,
        ignore_overlaps=ignore_overlaps,
        min_base_quality=min_base_quality,
        min_mapping_quality=min_mapping_quality,
    )

    for pileup in mpileup_iter:
        run_index = "%s_%s" % (chr_name, pileup.reference_pos + 1)

        if (pileup.reference_pos + 1) > region_end:
            break

        if site_dict.get(run_index) is not None:
            base_list = [i.upper() for i in pileup.get_query_sequences()]
            site_dict[run_index]["A"] = base_list.count("A")
            site_dict[run_index]["T"] = base_list.count("T")
            site_dict[run_index]["C"] = base_list.count("C")
            site_dict[run_index]["G"] = base_list.count("G")
            site_dict[run_index]["N"] = base_list.count("N")
            site_dict[run_index]["total"] = len(base_list)

    return site_dict


if __name__ == "__main__":
    pt_bam = "/Users/zhaohuanan/zhaohn_HD/3.project/2021_DdCBE_topic/20210224_DetectSeq_all_bams/bam/293T-DdCBE-ND6-All-PD_rep2_hg38.MAPQ20.bam"
    pt_ref = "/Users/zhaohuanan/zhaohn_HD/2.database/db_genomes/genome_fa/genome_ucsc_hg38/chr20.fa"

    # load bam
    bam_file = pysam.AlignmentFile(pt_bam, "rb")
    # load genome
    genome_ref = pysam.FastaFile(filename=pt_ref)
    get_query_site_mpileup_info(
        site_idx_list=["chr20_53025829", "chr20_53025830"],
        bam_obj=bam_file,
        genome_obj=genome_ref,
    )



import argparse
import logging
import sys

import numpy as np
import pandas as pd
import pysam

from loader import load_everybase_from_bowtie_table, load_reference_fasta_as_dict
from parse_pysam import get_query_site_mpileup_info

"""initialize"""
pd.set_option("max_colwidth", 60)  # column最大宽度
pd.set_option("display.width", 150)  # dataframe宽度
pd.set_option("display.max_columns", None)  # column最大显示数
pd.set_option("display.max_rows", 100)  # row最大显示数


def logging_setting():
    # log setting
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)-5s @ %(asctime)s: %(message)s ",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        filemode="w",
        force=True,
    )


def argparse_setting():
    # argparse setting
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--bowtie_table",
        type=str,
        required=True,
        help="""
            the result of: 
                bowtie \
                    path/genome.fa.bowtie1_index \
                    target-region.fa -f \
                    > target-region.align.tsv"
        """,
    )
    parser.add_argument(
        "--reference",
        type=str,
        required=True,
        help="""
            fasta, the same genome used to mapping bam file.
        """,
    )
    parser.add_argument(
        "--bam",
        type=str,
        required=True,
        help="""
            Such as Detect-seq Pull-down bam file.
            Must be sorted by position.
            Must have .bai index
        """,
    )
    parser.add_argument(
        "--out",
        type=str,
        default=None,
        help="""
            The path to put out csv file.
        """,
    )
    parser.add_argument(
        "--treatment",
        type=str,
        default=None,
        help="""
            Treatment of this sample
        """,
    )
    parser.add_argument(
        "--rep",
        type=str,
        default=None,
        help="""
            Replication of this sample: rep1, rep2
        """,
    )
    return parser.parse_args()


if __name__ == "__main__":
    # use logging
    logging_setting()
    # load params
    args = argparse_setting()
    pt_aligninfo = args.bowtie_table
    pt_ref = args.reference
    pt_bam = args.bam
    pt_out = args.out
    treatment = args.treatment
    rep = args.rep

    # load bam
    bam_file = pysam.AlignmentFile(pt_bam, "rb")
    # load ref obj
    genome_ref = pysam.FastaFile(filename=pt_ref)
    # load ref dict
    dt_ref = load_reference_fasta_as_dict(
        ref_fasta_path=pt_ref,
        ref_name="All",
        log_verbose=logging.DEBUG,
    )
    # load region info
    df_region = load_everybase_from_bowtie_table(
        bowtie_table=pt_aligninfo, narrow_cutoff=10, near_seq_extend=10
    )
    region_count = len(df_region.groupby("region_id"))
    now_count = 0

    dt_all_region = {}
    for idx_df_one_region, df_one_region in df_region.groupby("region_id"):
        now_count += 1
        ls_site_idx_all = df_one_region.site_idx.tolist()
        for i in range(0, len(ls_site_idx_all), 100):
            ls_site_idx = ls_site_idx_all[i : i + 100]

            dt_one_region = get_query_site_mpileup_info(
                site_idx_list=ls_site_idx,
                bam_obj=bam_file,
                genome_obj=genome_ref,
            )
            dt_all_region.update(dt_one_region)

        logging.info("Parsed %s/%s aim regions." % (now_count, region_count))

    df_detect = pd.DataFrame(dt_all_region).T
    df_detect.reset_index(inplace=True)
    df_detect.rename(columns={"index": "site_idx"}, inplace=True)
    df_detect_single_base = df_region.merge(df_detect, how="left", on=["site_idx"])

    if df_detect_single_base.isna().sum().sum() != 0:
        logging.fatal("merging file fail!!")
        raise ValueError("merging file fail!!")
    
    # 有BUG，不能用，先放着回头改！
    for col_base in df_detect_single_base[["A", "G", "C", "T"]]:
        df_detect_single_base["%s_ratio" % col_base] = (
            df_detect_single_base[col_base] / df_detect_single_base["total"]
        )
    if treatment is not None:
        treatment = treatment.strip()
        df_detect_single_base.insert(0, "treatment", treatment, allow_duplicates=False)
    else:
        df_detect_single_base.insert(0, "treatment", np.NaN, allow_duplicates=False)
    if rep is not None:
        df_detect_single_base.insert(1, "rep", rep, allow_duplicates=False)
    else:
        df_detect_single_base.insert(1, "rep", np.NaN, allow_duplicates=False)

    df_detect_single_base.to_csv(pt_out, header=True, index=False, sep=",")



import os
import time

def convert_size(size, mode='b') -> float:
    """将单位为Byte的数字转为其它单位

    :param size: int, 单位为Byte的数字
    :param mode: str, 单位，b、k、m、g、t或者B、K、M、G、T
    :return:
    """
    assert mode in list('bkmgtBKMGT')
    mode = mode.lower()

    if mode == 'b':
        return float(size)
    elif mode == 'k':
        return size / 1024.0
    elif mode == 'm':
        return size / 1024.0 / 1024.0
    elif mode == 'g':
        return size / 1024.0 / 1024.0 / 1024.0
    elif mode == 't':
        return size / 1024.0 / 1024.0 / 1024.0 / 1024.0
    else:
        return None


def get_size(path, mode='b'):
    """获取路径占用大小.
    通过递归函数获取路径占用大小
    :param path: str, 目标路径
    :param mode: str, 单位，b、k、m、g、t或者B、K、M、G、T
    :return: str, 占用大小
    """
    assert mode in list('bkmgtBKMGT')
    mode = mode.lower()

    # 目标是否存在
    if os.path.exists(path):
        # 目标路径是一个文件
        if os.path.isfile(path):
            # 无序遍历，直接返回即可
            byte_size = os.path.getsize(path)
            return convert_size(size=byte_size, mode=mode)

        # 目标路径是一个文件夹
        byte_size = 0.0
        files = os.listdir(path)

        for file in files:
            cur_path = os.path.join(path, file)
            try:
                byte_size += get_size(cur_path)
            except TypeError:
                continue

        return convert_size(size=byte_size, mode=mode)
    else:
        # print(f'{path} not exists!')
        pass


if __name__ == '__main__':
    # 大路径
    aim_paths = [
        # '/lustre1/chengqiyi_pkuhpc/zhaohn/2.disk_occupy/1T/fastq', # for test
        # '/lustre1/chengqiyi_pkuhpc/zhaohn/2.disk_occupy/1T.1/1T/fastq', # for test\
        # 
        '/lustre1/chengqiyi_pkuhpc/zhaohn',
        '/lustre1/chengqiyi_pkuhpc/JiangZhe',
        '/lustre1/chengqiyi_pkuhpc/LiuCong',
        '/lustre1/chengqiyi_pkuhpc/liuwenqing',
        '/lustre1/chengqiyi_pkuhpc/lubo',
        '/lustre1/chengqiyi_pkuhpc/shx',
        '/lustre2/chengqiyi_pkuhpc/bds',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/bds',
        '/lustre2/chengqiyi_pkuhpc/dyy',
        '/lustre2/chengqiyi_pkuhpc/LK',
        '/lustre2/chengqiyi_pkuhpc/log',
        '/lustre2/chengqiyi_pkuhpc/LSY',
        '/lustre2/chengqiyi_pkuhpc/RenweiLi',
        '/lustre2/chengqiyi_pkuhpc/wuhao',
        '/lustre2/chengqiyi_pkuhpc/xiongxs_tmp',
        '/lustre2/chengqiyi_pkuhpc/zhangxt',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/zhangxt',
        '/lustre2/chengqiyi_pkuhpc/zml',
        '/lustre3/chengqiyi_pkuhpc/Alumni/mshq',
        '/lustre3/chengqiyi_pkuhpc/Alumni/Shuxiaoting',
        '/lustre3/chengqiyi_pkuhpc/Alumni/SJH',
        '/lustre3/chengqiyi_pkuhpc/Alumni/WK',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/dlt',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/Leizhixin',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/LiuJiangle',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/lixy',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/lzc',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/test2',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/ydy',
        '/lustre3/chengqiyi_pkuhpc/folder_for_learning/ZhuangYuan',
    ]
    
    out_path = '/lustre1/chengqiyi_pkuhpc/zhaohn/__diskusage.all_user.csv'
    
    dt_size = {key: [] for key in aim_paths}
    
    for aim_path in aim_paths:
        file_sizes = []
        files = os.listdir(aim_path)
        
        # print("aim_path: ", aim_path)
        for file in files:
            # continue
            cur_path = os.path.join(aim_path, file)
            # print("cur_path: ", cur_path)
            if os.path.isdir(cur_path):
                try:
                    file_size = (file, get_size(cur_path, 't'))
                    file_sizes.append(file_size)
                    print(f"INFO:    Processed {file_size[0]}  {file_size[1]:.3f}TB")
                except Exception as err:  # 捕捉所有异常
                    print(f"WARNING: {err} @ {cur_path}, Skip ...")
                except FileNotFoundError:
                    print(f"FileNotFoundError @ {cur_path}")
                    continue
                except Exception as err:
                    print(f"Exception @ {cur_path}: {err}")
                    
        # print("file_sizes: ", file_sizes)
        file_sizes.sort(key=lambda t: t[1], reverse=True)
        dt_size[aim_path] = file_sizes
    date_time = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime())
        
    with open(out_path, 'wt') as f:
        f.write(f'{date_time},by,Huanan Zhao,\n,,,\n')
        f.write("USER,PATH,TOTAL_SIZE,UNIT\n")        
        # print(dt_size)
        
        for key in dt_size.keys():
            user = os.path.basename(key)
            size = sum([item[1] for item in dt_size[key]])
            f.write(f'{user},{key},{size:>3f},TB\n')

    print(f'one loop done! @ {date_time}')



import logging
import sys

import pandas as pd
from Bio import Seq, SeqIO


def load_reference_fasta_as_dict(
    ref_fasta_path, ref_name="All", log_verbose=logging.DEBUG
) -> dict:
    """Load the reference fasta file and return a dict of reference.

    Args:
        ref_fasta_path (str):
            Reference fasta file path
        ref_name (str/list, optional):
            If set All, load all "seq ids" in reference,
            else only try to load specific chromosome name, such as "chr1". Defaults to "All".
        log_verbose (logging object, optional): a logging states object, which can be logging.DEBUG,
            logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL. Defaults to logging.DEBUG.

    Raises:
        IOError:
            Load file error

    Returns:
        dict/None:
            Normally, this function returns a dict, the key is the right "seq id" in the ref_name list and the value is sequences with .upper();
            it returns a empty_dict when there is no right "seq id" in the ref_name list;
            the loss of the key_value_pair can coursed by the wrong "seq id";
            it returns None when error occurs.
    """
    # log setting
    logging.basicConfig(
        level=log_verbose,
        format="%(levelname)-5s @ %(asctime)s: %(message)s ",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        filemode="w",
        force=True,
    )
    # load genome as dict
    try:
        genome_fa = SeqIO.parse(handle=ref_fasta_path, format="fasta")
    except OSError:
        raise OSError("Wrong file path: %s" % ref_fasta_path)

    # init var
    ref_seq_dict = {}
    ref_name_set = set(ref_name)

    logging.info("Starting to load the reference genome...")

    for ref in genome_fa:
        if ref_name == "All":
            ref_seq_dict[ref.id] = ref.seq.upper()
            logging.debug("Loading genome...\t" + ref.id)

        elif ref.id in ref_name:
            ref_seq_dict[ref.id] = ref.seq.upper()
            logging.debug("Loading genome...\t" + ref.id)

            # remove already loaded seq
            ref_name_set.remove(ref.id)

            # load all info
            if len(ref_name_set) == 0:
                break

    logging.info("Loading genome done!")

    return ref_seq_dict


def get_reference_sequence_from_aligned_segment(align, ref_genome_dict) -> SeqIO:
    """get reference sequence of pysam.AlignedSeqment object

    Args:
        align (pysam.AlignedSeqment object):
            pysam.AlignedSeqment object
        ref_genome_dict (dict, optional):
            returned dict from load_reference_fasta_as_dict().
    Returns:
        Bio.Seq object:
            it returns reference sequence (Bio Seq object)
    """
    MD_tag_state = align.has_tag("MD")

    # align_start_idx = align.get_aligned_pairs()[0][0]
    ref_start_idx = align.get_aligned_pairs()[0][1]
    # align_end_idx = align.get_aligned_pairs()[-1][0]
    ref_end_idx = align.get_aligned_pairs()[-1][1]

    if MD_tag_state:
        return align.get_reference_sequence().__str__()
    else:
        return ref_genome_dict[align.reference_name][ref_start_idx : ref_end_idx + 1]


def load_region_absolute_position(bowtie_table: str) -> pd.DataFrame:
    """get region position from bowtie1 mapping standard output file

    Args:
        bowtie_table (str): path of standard output file (must use the same reference with load_reference_fasta_as_dict())

    Returns:
        pd.DataFrame: the absolution positions of aimed regions.
        cols: [
            "chrom",  # chromosome
            "start",  # start (contain)
            "stop",  # stop (contain)
            "region_id",  # region name
            "score",  # fake score
            "strand",  # strand
            "sequence",  # sequence (ref strand +)
            ]
    """
    df = pd.read_csv(
        bowtie_table,
        sep="\t",
        names=[
            "region_id",
            "strand",
            "chrom",
            "start",
            "sequence",
            "align_info",
            "score",
        ],
        index_col=False,
    )
    print(df[df.region_id == "ND6-only-2(+)"].sequence.values)
    df.insert(4, "stop", df["start"] + df.sequence.map(len), allow_duplicates=False)
    df.region_id = df.region_id.map(lambda x: x.split("(")[0])
    df.start = df.start + 1
    return df[["chrom", "start", "stop", "region_id", "score", "strand", "sequence"]]


def load_everybase_from_bowtie_table(
    bowtie_table: str,
    # narrow_cutoff=10,
    # near_seq_extend=10,
    narrow_cutoff=0,
    near_seq_extend=0,
) -> pd.DataFrame:
    if narrow_cutoff < near_seq_extend:
        raise ValueError("Error: narrow_cutoff must >= near_seq_extend")

    df = load_region_absolute_position(bowtie_table)

    ls = []

    for idx, df_row in df.iterrows():
        region_id = df_row["region_id"]
        chrom = df_row["chrom"]
        start = df_row["start"] + 1
        strand = df_row["strand"]
        lens = len(df_row["sequence"])
        genome_seq = df_row["sequence"]

        for idx_seq, genome_base in enumerate(genome_seq):
            if idx_seq < narrow_cutoff:
                continue
            elif idx_seq > lens - narrow_cutoff:
                continue

            if strand == "+":
                relative_pos = idx_seq + 1
                near_seq = Seq.Seq(
                    genome_seq[idx_seq - near_seq_extend : idx_seq + near_seq_extend]
                )
                region_seq = Seq.Seq(genome_seq)
            elif strand == "-":
                relative_pos = lens - idx_seq
                near_seq = Seq.Seq(
                    genome_seq[idx_seq - near_seq_extend : idx_seq + near_seq_extend]
                ).complement()
                genome_base = Seq.Seq(genome_base).complement()
                region_seq = Seq.Seq(genome_seq).complement()
            absolute_pos = start + idx_seq - 1
            ls.append(
                [
                    region_id,
                    chrom,
                    genome_base,
                    relative_pos,
                    absolute_pos,
                    strand,
                    region_seq,
                    near_seq,
                ]
            )
    df_out = pd.DataFrame(
        ls,
        columns=[
            "region_id",
            "chrom",
            "genome_base",
            "relative_pos",
            "absolute_pos",
            "strand",
            "region_seq",
            "near_seq",
        ],
    )
    df_out["site_idx"] = df_out.chrom + ["_"] + df_out.absolute_pos.map(str)
    df_out["sort_chrom"] = df_out.chrom.map(
        lambda x: x.replace("chr", "")
        .replace("X", "23")
        .replace("Y", "24")
        .replace("M", "25")
    ).map(int)
    df_out.sort_values(
        by=["sort_chrom", "absolute_pos"], inplace=True, ignore_index=True
    )
    return df_out


# unit test
if __name__ == "__main__":
    # pt_aligninfo = "Bed2DetectSeqInfo/test.align.tsv"
    pt_aligninfo = "./test.align.tsv"
    df = load_region_absolute_position(pt_aligninfo)
    df[df.region_id == "ND6-only-2"]
    len((df[df.region_id == "ND6-only-2"].sequence).values.tolist()[0])
    df.loc[df.region_id == "ND6-only-2", "sequence"].values
    df2 = load_everybase_from_bowtie_table(pt_aligninfo)
    df2[df2.region_id == "ND6-only-2"]
    df.loc[df.region_id == "ND6-only-2", "sequence"].values



import argparse
import os
import sys

import pysam


class RemoveClip():
    pass


def compared_version(ver1, ver2):
    """Compared_version number.
    Compared_version number like: "9.1.1">"10.1.2", or "10.12.2.6.5">"10.12.2.6".

    Args:
        ver1: version 1
        ver2: version 2
    Returns:
        int, -1/0/1, ver1<ver2, return -1; ver1==ver2; return 0; ver1 > ver2, return 1.
    """
    list1 = str(ver1).split(".")
    list2 = str(ver2).split(".")
    # 循环次数为短的列表的len
    for i in range(len(list1)) if len(
            list1) < len(list2) else range(len(list2)):
        if int(list1[i]) == int(list2[i]):
            pass
        elif int(list1[i]) < int(list2[i]):
            return -1
        else:
            return 1
    # 循环结束，哪个列表长哪个版本号高
    if len(list1) == len(list2):
        return 0
    elif len(list1) < len(list2):
        return -1
    else:
        return 1


def check_version():
    if compared_version(pysam.__version__, "0.19.1") == -1:
        raise ValueError(
            f"pysam version is {pysam.__version__} and it should >=0.19.1")
    if compared_version(sys.version.split()[0], "3.9.1") == -1:
        raise ValueError(
            f"Python version is {sys.version.split()[0]} and it should >= 3.9.1:)")


def get_args():
    parser = argparse.ArgumentParser(
        description="remove softclip reads in BAM file")

    parser.add_argument(
        "-i",
        "--input",
        help="BAM file sorted by coordinate with soft/hard clip reads",
        type=str,
        nargs='?',
        default=sys.stdin)

    parser.add_argument(
        "-o",
        "--output",
        help="BAM file sorted by coordinate without soft/hard clip reads",
        type=str,
        nargs='?',
        default=sys.stdout)
    parser.add_argument(
        "-t",
        "--threads",
        help="threads used by pysam and samtools core, default=os.cpu_count() - 1",
        type=str,
        default=os.cpu_count() -
                1)
    parser.add_argument(
        "-b",
        "--output_format",
        help="BAM, SAM(default)",
        type=str,
        default="SAM"
    )
    parser.add_argument(
        "-p",
        "--remove_as_paired",
        help="True, False. remove single/paired-reads if any clip in read1 or read2(default: True)",
        type=bool,
        default=True)
    parser.add_argument(
        "-c",
        "--max_clip",
        help="the maximum clips allowed per read, default=0",
        type=int,
        default=0
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = get_args()

    # [print(i) for i in dir(pysam)]

    # input_bam = "../Test_Resources/test_sorted.bam"
    # # input_bam = "../Test_Resources/test_sorted_n.bam"
    # output_bam = "../Test_Outfiles/test_sorted_rmclip.bam"
    check_version()

    if sys.stdin.isatty():
        try:
            bam_in = pysam.AlignmentFile(args.input, "r", threads=args.threads, check_sq=False)
        except ValueError:
            bam_in = pysam.AlignmentFile(args.input, "rb", threads=args.threads, check_sq=False)
    else:
        try:
            bam_in = pysam.AlignmentFile("-", "r", threads=args.threads, check_sq=False)
        except ValueError:
            bam_in = pysam.AlignmentFile("-", "rb", threads=args.threads, check_sq=False)

    try:
        so = bam_in.header['HD']['SO']
    except KeyError:
        so = None

    if so != "queryname" or so == "coordinate" or so is None:
        raise ValueError(f"the input BAM|SAM must be sorted by name and has header [SO:queryname]!\n"
                         f"your header: [SO:{so}]\n")

    write_mode = 'wb' if args.output_format == "BAM" else 'w'
    bam_out = pysam.AlignmentFile(args.output, "w", template=bam_in)

    # Iterate through reads.
    read1, read2 = None, None

    for read in bam_in:
        if not read.is_paired or read.is_unmapped or read.mate_is_unmapped:
            # or read.is_duplicate:
            # only use mapped paired reads
            continue
        # ------------------------------------------------------------------->>>>>>>>>>
        # 如果先看到read1，将read1赋给read， 判断是否read1、read2都存在
        #    不都存在：
        #        即read2不存在，进入下个循环
        #    都存在，判断query name是否一致
        # ------------------------------------------------------------------->>>>>>>>>>
        if read.is_read2:
            # if first is read2, then use it to find read2
            read2 = read
        else:
            # if first is read1, then use it to find read1, and next query
            read1 = read
            read2 = None
            continue

        if None not in [
            read1,
            read2] and read1.query_name == read2.query_name:
            # print("found a pair!")
            # print(f'Cigar: Read1-{read1.cigarstring}, Read2-{read1.cigarstring}')
            # print(read.get_cigar_stats())
            softclip1 = read1.get_cigar_stats()[0][4]
            softclip2 = read2.get_cigar_stats()[0][4]

            if not args.remove_as_paired:
                if softclip1 <= args.max_clip:
                    bam_out.write(read1)
                if softclip2 <= args.max_clip:
                    bam_out.write(read2)
            elif args.remove_as_paired:
                if softclip1 <= args.max_clip and softclip2 <= args.max_clip:
                    bam_out.write(read1)
                    bam_out.write(read2)
                else:
                    # print("Filtering pairs")
                    # print(f'Cigar: Read1-{read1.cigarstring}, Read2-{read2.cigarstring}')
                    # print(read1.get_cigar_stats()[0][4])
                    # print(read2.get_cigar_stats()[0][4])
                    pass
            else:
                pass
                # print(f'Cigar: Read1-{read1.cigarstring}, Read2-{read2.cigarstring}')
                # raise ValueError(
                #     f"remove_se_or_pe can not be {args.remove_se_or_pe}")

    bam_in.close()
    bam_out.close()



import argparse
import gzip
import os
import random
import string
import sys
import time
from multiprocessing import Process

import numpy as np


# ------------------------------------------------------------------->>>>>>>>>>
# define params
# ------------------------------------------------------------------->>>>>>>>>>
def get_parser():
    """get params from cmd python bmatFilter.py -<item> [parma]

    Returns:
        object: Object instance of argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        description="filter of bmat/bmat.gz, filter param全是与或非中的与关系，也就是条件同时成立！"
    )
    parser.add_argument(
        "-i", "--Input", required=True, type=str, help="输入bmat或者bmat.gz"
    )
    parser.add_argument(
        "-o", "--Output", required=True, type=str, help="输出bmat或者bmat.gz"
    )
    parser.add_argument(
        "-p",
        "--Threads",
        default=1,
        type=int,
        help="Multiple threads number, default=1",
    )
    parser.add_argument(
        "--TempDir",
        help="Where to keep temp files, default is the same dir with --Input",
        default=None,
    )
    parser.add_argument(
        "--MinDepth",
        default=0,
        type=int,
        help="能够接受的单碱基上覆盖的最小read count数，default=0不做过滤",
    )
    parser.add_argument(
        "--MinMutNum",
        default=-1,
        type=int,
        help=""""能够接受的单碱基最小突变数，
        小于MinMutNum的位点会被过滤；
        如果MinMutNum=0，则输出所有位点；
        如果MinMutNum=1，则不含突变的位点会被过滤；
        如果MinMutNum=2，则含有0个、1个突变的位点会被过滤。default=-1输出所有位点（和0作用一致）"
        """,
    )
    parser.add_argument(
        "--MaxMutNum",
        default=-1,
        type=int,
        help="""能够接受的单碱基最大突变数, 大于MaxMutNum的位点会被过滤；
        如果MaxMutNum=1，则只输出无突变的位点；
        如果MaxMutNum=2，则输出含有0个、1个突变的位点；
        如果MaxMutNum=-1，则不进行过滤。default=-1。
        default=-1，则不使用该过滤参数
        """,
    )
    parser.add_argument(
        "--MinMutRatio",
        default=-1,
        type=float,
        help="""能够接受的单碱基最小突变比率，default=-1，则不使用该过滤参数""",
    )
    parser.add_argument(
        "--MaxMutRatio",
        default=-1,
        type=float,
        help="""能够接受的单碱基最小突变比率，default=-1，则不使用该过滤参数""",
    )
    return parser


# ------------------------------------------------------------------->>>>>>>>>>
# split file for multiprocess
# ------------------------------------------------------------------->>>>>>>>>>
def split_file_and_make_temp(input_filename, threads_num=1, temp_dir=None):
    """split input file for multi threads calculating

    Args:
        input_filename (str): path of input file
        threads_num (int, optional): number of cores to use. Defaults to 1.
        temp_dir (str, optional): path of splited temp files. Defaults to None.

    Returns:
        list: path list of split files
    """
    # --------------------------------------------------->>>>>>
    # set temp dir
    # --------------------------------------------------->>>>>>
    if temp_dir is None:
        temp_dir = os.path.dirname(input_filename)

    # --------------------------------------------------->>>>>>
    # get input basename
    # --------------------------------------------------->>>>>>
    input_file_basename = os.path.basename(input_filename)

    # --------------------------------------------------->>>>>>
    # creat temp file name and open file
    # --------------------------------------------------->>>>>>
    # >>>>>> LOG
    sys.stderr.write(
        "Making temp files... \t %s \n"
        % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )

    temp_filename_list = []
    temp_file_list = []

    for index in range(threads_num):
        # make filename
        temp_file_basename = (
            "temp_"
            + input_file_basename
            + "."
            + str(index)
            + "."
            + "".join(random.sample(string.ascii_letters + string.digits, 16))
        )
        temp_file_name = os.path.join(temp_dir, temp_file_basename)
        temp_filename_list.append(temp_file_name)

        # open file
        if IfGzipInput:
            temp_file = gzip.open(temp_file_name, "w")
        elif not IfGzipInput:
            temp_file = open(temp_file_name, "w")
        temp_file_list.append(temp_file)

    # --------------------------------------------------->>>>>>
    # counting input file line number
    # --------------------------------------------------->>>>>>
    # >>>>>> LOG
    sys.stderr.write(
        "Counting input file... \t %s \n"
        % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )

    if IfGzipInput:
        input_file = gzip.open(input_filename, "r")
    elif not IfGzipInput:
        input_file = open(input_filename, "r")

    total_input_line_num = 0
    for line in input_file:
        total_input_line_num += 1

    if total_input_line_num % threads_num == 0:
        each_file_line_num = total_input_line_num // threads_num
    else:
        each_file_line_num = (total_input_line_num // threads_num) + 1

    input_file.close()

    # >>>>>> LOG
    sys.stderr.write(
        "Done! \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )

    # --------------------------------------------------->>>>>>
    # split temp files
    # --------------------------------------------------->>>>>>
    # write into output filename
    if IfGzipInput:
        input_file = gzip.open(input_filename, "r")
    else:
        input_file = open(input_filename, "r")

    for index, line in enumerate(input_file):
        file_index = index // each_file_line_num
        temp_file_list[file_index].write(line)

    # close output filename
    input_file.close()
    [temp_file.close() for temp_file in temp_file_list]

    # >>>>>> LOG
    sys.stderr.write(
        "Make temp files done! \t %s \n"
        % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )

    return temp_filename_list


# ------------------------------------------------------------------->>>>>>>>>>
# merge temp to output
# ------------------------------------------------------------------->>>>>>>>>>
def merge_out_bmat(temp_filename_list, output_filename, remove_temp=True):
    """merge temp files to output file

    Args:
        temp_filename_list (list): the returned list of func: split_file_and_make_temp
        output_filename (str): path of output file
        remove_temp (bool, optional): Whether to delete the temp files after the program runs. Defaults to True.
    """

    # write header
    header = [
        "chr_name",
        "chr_index",
        "ref_base",
        "A",
        "G",
        "C",
        "T",
        "del_count",
        "insert_count",
        "ambiguous_count",
        "deletion",
        "insertion",
        "ambiguous",
        "mut_num",
    ]

    # open output file
    if output_filename == "Stdout":
        output_file = sys.stdout
    elif IfGzipOutput:
        output_file = gzip.open(output_filename, "w")
    elif not IfGzipOutput:
        output_file = open(output_filename, "w")

    # write header
    if IfGzipOutput:
        output_file.write(("\t".join(header) + "\n").encode("utf-8"))
    elif not IfGzipOutput:
        output_file.write("\t".join(header) + "\n")

    # open temp bmat file and write to merge file
    for temp_filename in temp_filename_list:
        # write file
        if IfGzipInput:
            with gzip.open(temp_filename + ".bmat", "r") as temp_file:
                for line in temp_file:
                    output_file.write(line)
        elif not IfGzipInput:
            with open(temp_filename + ".bmat", "r") as temp_file:
                for line in temp_file:
                    output_file.write(line)
    # remove temp files
    if remove_temp:
        for temp_filename in temp_filename_list:
            os.remove(temp_filename)
            os.remove(temp_filename + ".bmat")


# ------------------------------------------------------------------->>>>>>>>>>
# filter core function
# ------------------------------------------------------------------->>>>>>>>>>
def core_filter(
    Input,
    Output="Stdout",
    IfGzipInput=False,
    IfGzipOutput=False,
    MinDepth=0,
    MinMutNum=-1,
    MaxMutNum=-1,
    MinMutRatio=-1,
    MaxMutRatio=-1,
):
    """the filter function to screen bmat lines

    Args:
        Input (str): one splited path name from the returned list of function: split_file_and_make_temp
        Output (str, optional): path of output file. Defaults to "Stdout".
        MinDepth (int, optional): min read count of one base. Defaults to 0.
        MinMutNum (int, optional): . Defaults to 0.
    """
    # open input file
    if IfGzipInput:
        InputFile = gzip.open(Input, "r")
    elif not IfGzipInput:
        InputFile = open(Input, "r")
    # open output file
    if Output == "Stdout":
        OutputFile = sys.stdout
    elif IfGzipOutput:
        OutputFile = gzip.open(Output, "a")
    elif not IfGzipOutput:
        OutputFile = open(Output, "a")

    # iterator
    for line in InputFile:
        if IfGzipInput:
            line = line.decode("utf-8")
        elif not IfGzipInput:
            pass

        ls_line = line.strip().split("\t")

        # 不分析header行
        if ls_line[0] == "chr_name":
            continue

        array_line = np.array([int(i) for i in ls_line[3:7]])
        LineDepth = np.sum(array_line)

        # 不满足最小深度的，过滤掉
        if LineDepth < MinDepth:
            continue

        MutCount = int(ls_line[-1])
        # 小于可接受最小突变数的，过滤掉
        if (MinMutNum != -1) and (MutCount < MinMutNum):
            continue
        # 大于可接受最大突变数的，过滤掉
        if (MaxMutNum != -1) and (MutCount > MaxMutNum):
            continue

        # ratio和Num一起作用
        ThisBaseMutRatio = MutCount / LineDepth if LineDepth > 0.0 else 0.0
        # 小于可接受最小突变率的，过滤掉
        if (MinMutRatio != -1.0) and (ThisBaseMutRatio < MinMutRatio):
            continue
        # 大于可接受最大突变率的，过滤掉
        if (MaxMutRatio != -1.0) and (ThisBaseMutRatio > MaxMutRatio):
            continue

        # ------------------------------------------------------------------->>>>>>>>>>
        # 临时条件
        # if($total <= 100 && $total >= 50){
        #     next if $mutations < ($total-10);#mismatch < depth - 10 # == if True: continue
        #     print OUT "$_\n";
        # }
        # elsif($total > 100){
        #     next if $mutations < ($total-20);#mismatch < depth - 20 # == if True: continue
        #     print OUT "$_\n";
        # ------------------------------------------------------------------->>>>>>>>>>
        # if LineDepth < 50:
        #     continue
        # if (LineDepth <= 100) and (LineDepth >= 50):
        #     if MutCount < LineDepth - 10:
        #         continue
        # elif LineDepth > 100:
        #     if MutCount < LineDepth - 20:
        #         continue
        # ------------------------------------------------------------------->>>>>>>>>>

        # 通过 filter 的行被写入输出文件
        str_out = "\t".join(ls_line)
        str_out = str_out + "\n"
        if IfGzipOutput:
            OutputFile.write(str_out.encode("utf-8"))
        elif not IfGzipOutput:
            OutputFile.write(str_out)
    InputFile.close()
    OutputFile.close()


if __name__ == "__main__":
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # load the paramters
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    parser = get_parser()
    args = parser.parse_args()
    InputFilePath = args.Input
    OutputFilePath = args.Output
    Threads = args.Threads
    TempDir = args.TempDir
    MinDepth = args.MinDepth
    MinMutNum = args.MinMutNum
    MaxMutNum = args.MaxMutNum
    MinMutRatio = args.MinMutRatio
    MaxMutRatio = args.MaxMutRatio
    print("*" * 30)
    print("MinDepth\t", MinDepth)
    print("MinMutNum\t", MinMutNum)
    print("MaxMutNum\t", MaxMutNum)
    print("MinMutRatio\t", MinMutRatio)
    print("MaxMutRatio\t", MaxMutRatio)
    print("*" * 30)
    IfGzipInput = True if InputFilePath[-3:] == ".gz" else False
    IfGzipOutput = True if OutputFilePath[-3:] == ".gz" else False
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # remove output file
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    if os.path.exists(OutputFilePath):
        print("file %s exists & remove it" % OutputFilePath)
        os.remove(OutputFilePath)
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # make temp files
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    temp_filename_list = split_file_and_make_temp(
        input_filename=InputFilePath, threads_num=Threads, temp_dir=TempDir
    )
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    # multiple processing part
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    sys.stderr.write(
        "Parsing files... \t %s \n"
        % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )

    # multiprocess
    procs_list = []

    for index, temp_filename in enumerate(temp_filename_list):
        out_filename = temp_filename + ".bmat"

        # add multiprocess
        # params

        sub_proc = Process(
            target=core_filter,
            args=(
                temp_filename,  # Input,
                out_filename,  # Output="Stdout",
                IfGzipInput,  # IfGzipInput=False,
                IfGzipOutput,  # IfGzipOutput=False,
                MinDepth,  # MinDepth=0,
                MinMutNum,  # MinMutNum=-1,
                MaxMutNum,  # MaxMutNum=-1,
                MinMutRatio,  # MinMutRatio=-1,
                MaxMutRatio,  # MaxMutRatio=-1,
            ),
        )
        procs_list.append(sub_proc)
        sub_proc.start()

    for sub_proc in procs_list:
        sub_proc.join()

    sys.stderr.write(
        "Done! \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )
    # >>>>>> LOG
    sys.stderr.write(
        "Merging files... \t %s \n"
        % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )

    merge_out_bmat(
        temp_filename_list=temp_filename_list,
        output_filename=OutputFilePath,
        remove_temp=True,
    )

    # >>>>>> LOG
    sys.stderr.write(
        "Done!... \t %s \n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    )



#!/usr/bin/env python
from bioat import __version__
from cleo.application import Application as BaseApplication
from bioat.console.commands.bam_tools import BamToolsCommand
from bioat.console.commands.bed_tools import BedToolsCommand


class Application(BaseApplication):
    def __init__(self) -> None:
        super().__init__(name="bioat", version=__version__)

        for command in self.get_default_commands():
            self.add(command)

    def get_default_commands(self) -> list:
        commands = [
            BamToolsCommand(),
            BedToolsCommand(),
        ]
        return commands


def main() -> int:
    application = Application()
    exit_code: int = application.run()
    return exit_code


if __name__ == "__main__":
    main()



#! /Users/meng/menghw_HD/miniconda3/bin/python
# _*_ coding: UTF-8 _*_

import argparse
import sys

import matplotlib
import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection

matplotlib.use('Agg')
np.set_printoptions(suppress=True)

# Version information START --------------------------------------------------
VERSION_INFO = \
    """

Author: MENG Howard

Version-02:
    2022-02-13
        Support Python3

Version-01:
    2021-05-01
        Make plot for DdCBE TALE sequence

    2019-12-05 
        Plot Alignment Result Table (.art) files

E-Mail: meng_howard@126.com

"""
# Version information END ----------------------------------------------------


# Learning Part START --------------------------------------------------------
LEARNING_PART = \
    """

There will be a lot of parameters to set,

but default parameters can handle output figure in most cases.

"""

PIPELINE = \
    """

1. Load table 

2. sort table with key words

3. make plot

"""
# Learning Part END-----------------------------------------------------------

###############################################################################
# main part
###############################################################################
if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Plot (.art) file")

    parser.add_argument("-i", "--input_art",
                        help="Input alignment result table", required=True)

    parser.add_argument("--align_seq",
                        help="Alignment sequence", required=True)

    parser.add_argument("-k", "--sort_column_key",
                        help="Sort .art file by this columns order, use ',' as separator, default=None means don't "
                             "sort .art file",
                        default=None)

    parser.add_argument("-r", "--sort_ascend_state",
                        help="Ture means sort by ascending order, False means sort by descending order, "
                             "default=False,...",
                        default=None)

    parser.add_argument("-a", "--annotation_column_key",
                        help="The annotation info in output figure are default set as "
                             "align_total_mismatch,"
                             "align_degen_total_mismatch,"
                             "region_index,"
                             "align_coordinate,"
                             "align_strand",
                        default="align_total_mismatch,"
                                "align_degen_total_mismatch,"
                                "region_index,"
                                "align_coordinate,"
                                "align_strand")

    parser.add_argument("-o", "--out_figure",
                        help="Output figure filename", required=True)

    parser.add_argument("--out_figure_format",
                        help="Support 'pdf' and 'png' result, default=pdf", default="pdf")

    parser.add_argument("--out_figure_dpi",
                        help="Out figure dpi, default=96", default="96", type=int)

    ARGS = parser.parse_args()

    # ------------------------------------------------------------------------>>>>>>>
    # load sgRNA
    # ------------------------------------------------------------------------>>>>>>>
    sgRNA_full_seq = str(ARGS.align_seq).upper()
    sgRNA_model_seq = sgRNA_full_seq

    # ---------------------------------------------------------------->>>>>>>>
    # load art 
    # ---------------------------------------------------------------->>>>>>>>
    art_table = pd.read_csv(ARGS.input_art, sep="\t")

    # make coordinate
    align_coordinate_list = []
    for index in art_table.index:
        chr_name, align_start, align_end = art_table.loc[
            index, ["align_chr_name", "align_chr_start", "align_chr_end"]]
        align_coordinate_str = "{chr_name}:{chr_start}-{chr_end}".format(chr_name=chr_name,
                                                                         chr_start=align_start,
                                                                         chr_end=align_end)
        align_coordinate_list.append(align_coordinate_str)
    art_table["align_coordinate"] = align_coordinate_list

    # ---------------------------------------------------------------->>>>>>>>
    # sort art
    # ---------------------------------------------------------------->>>>>>>>
    if ARGS.sort_column_key is not None:
        sort_key_list = ARGS.sort_column_key.split(",")
        sort_key_check_state = all([x in art_table.columns for x in sort_key_list])

        if sort_key_check_state:
            if ARGS.sort_ascend_state is None:
                sort_ascending_list = [False] * len(sort_key_list)

            else:
                try:
                    sort_ascending_list = []
                    for sort_state_str in ARGS.sort_ascend_state.split(","):
                        if sort_state_str == "False":
                            sort_ascending_list.append(False)
                        else:
                            sort_ascending_list.append(True)

                except:
                    sys.stderr.write("WARNING:\n--sort_ascend_state format error, use default action!\n")
                    sort_ascending_list = [False] * len(sort_key_list)

                if len(sort_ascending_list) != len(sort_key_list):
                    sys.stderr.write("WARNING:\n--sort_ascend_state format error, use default action!\n")
                    sort_ascending_list = [False] * len(sort_key_list)

        art_table_filter = art_table.sort_values(by=sort_key_list, ascending=sort_ascending_list)

    else:
        art_table_filter = art_table

    # ---------------------------------------------------------------->>>>>>>>
    # annotation list
    # ---------------------------------------------------------------->>>>>>>>
    ann_colname_list = list(ARGS.annotation_column_key.split(","))

    ann_colname_check_state_list = []

    for ann_colname in ann_colname_list:
        if ann_colname in art_table.columns:
            ann_colname_check_state_list.append(True)
        else:
            ann_colname_check_state_list.append(False)

    ann_colname_check_state = all(ann_colname_check_state_list)

    if not ann_colname_check_state:
        sys.stderr.write("WARNING:\n--annotation_column_key format error, use default action!\n")
        ann_colname_list = ["align_total_mismatch",
                            "align_degen_total_mismatch",
                            "region_index",
                            "align_coordinate",
                            "align_strand"]

    # ---------------------------------------------------------------->>>>>>>>
    # plot region
    # ---------------------------------------------------------------->>>>>>>>

    # --------------------------------------------->>>>>
    # set all box position
    # --------------------------------------------->>>>>
    # set plot column num 
    box_column_num = len(sgRNA_model_seq) + len(ann_colname_list)
    box_row_num = len(art_table_filter) + 1

    panel_box_width = 0.3
    panel_box_height = 0.3
    panel_row_space = 0.03
    panel_col_space = 0.03

    panel_box_width_coef = [1] * len(sgRNA_model_seq)

    for colname in ann_colname_list:
        col_info_len_list = [len(str(col_info)) for col_info in art_table_filter[colname]]
        col_info_len_list.append(len(colname))
        col_width_coef = max(col_info_len_list) // 3 + 1
        panel_box_width_coef.append(col_width_coef)

    # set figure width and height 
    figure_width = panel_box_width * sum(panel_box_width_coef) + panel_col_space * (box_column_num - 1)
    figure_height = panel_box_height * box_row_num + panel_row_space * (box_row_num - 1)

    # calculate all box x vec
    box_x_left_vec = []
    box_x_right_vec = []

    x_left = x_right = 0
    for index, col_coef in enumerate(panel_box_width_coef):
        if index > 0:
            x_left += panel_box_width_coef[index - 1] * panel_box_width + panel_col_space
            x_right = x_left + panel_box_width_coef[index] * panel_box_width
        else:
            x_left = 0
            x_right = x_left + panel_box_width_coef[0] * panel_box_width

        box_x_left_vec.append(x_left)
        box_x_right_vec.append(x_right)

    # fix number 
    box_x_left_vec = np.array(box_x_left_vec)
    box_x_right_vec = np.array(box_x_right_vec)

    box_x_left_vec.round(decimals=5)
    box_x_right_vec.round(decimals=5)

    # --------------------------------------------->>>>>
    # plot alignment region
    # --------------------------------------------->>>>>
    box_border_plot_state = False

    # make box border 
    if box_border_plot_state:
        box_edgecolor = "#AAAAAA"
        box_linestyle = "-"
        box_linewidth = 2
    else:
        box_edgecolor = "#FFFFFF"
        box_linestyle = "None"
        box_linewidth = 0

    # base_color_dict = {"A": "#04E3E3", "T": "#F9B874", "C": "#B9E76B", "G": "#F53798", "N": "#AAAAAA", "-": "#AAAAAA"}
    base_color_dict = {"A": "#97C3E9", "T": "#F7B6B7", "C": "#A2DB69", "G": "#D868B9", "N": "#DBDBDB", "-": "#DBDBDB"}

    # set new figure 
    fig = plt.figure(figsize=(figure_width * 1.1, figure_height * 1.1))
    ax = fig.add_subplot(111, aspect="equal")
    plt.xlim([0, figure_width])
    plt.ylim([-figure_height, 0])
    plt.axis("off")

    # make plot 
    text_list = []
    patches = []

    current_y = 0
    # ---------------------------------------->>>>>
    # plot model and colname 
    # ---------------------------------------->>>>>
    # plot model
    col_index = 0
    box_y_btm = current_y - panel_box_height

    for index, rna_char in enumerate(sgRNA_model_seq):
        box_x_left = box_x_left_vec[index]
        box_width = box_x_right_vec[index] - box_x_left_vec[index]
        box_value = rna_char

        box_color = base_color_dict.get(box_value)
        if not box_color:
            box_fill = False
            box_color = "#FFFFFF"
        else:
            box_fill = True

        # plot sgRNA box
        patches.append(Rectangle(
            xy=(box_x_left, box_y_btm),
            width=panel_box_width_coef[index] * panel_box_width,
            height=panel_box_height,
            fill=box_fill,
            alpha=1,
            linestyle=box_linestyle,
            linewidth=box_linewidth,
            edgecolor=box_edgecolor,
            facecolor=box_color)
        )

        # add sgRNA text
        text_list.append((box_x_left + 0.5 * box_width, box_y_btm + 0.5 * panel_box_height, box_value, 16))

        # record index 
        col_index += 1

        # plot name
    for ann_colname in ann_colname_list:
        box_x_left = box_x_left_vec[col_index]
        box_width = box_x_right_vec[col_index] - box_x_left_vec[col_index]
        text_list.append((box_x_left + 0.5 * box_width, box_y_btm + 0.5 * panel_box_height, ann_colname, 10))
        col_index += 1

        # make cur_y
    current_y = current_y - panel_box_height - panel_row_space

    # ---------------------------------------->>>>>
    # plot remain info
    # ---------------------------------------->>>>>
    for index in art_table_filter.index:
        sgRNA_align, align_info, query_align = art_table_filter.loc[
            index, ["align_query_seq", "align_info_state", "align_target_seq"]]
        ann_value_list = list(art_table_filter.loc[index, ann_colname_list])

        col_index = 0
        sgRNA_gap_count = 0
        box_y_btm = current_y - panel_box_height

        for index, rna_char in enumerate(sgRNA_align):
            if rna_char != "-":
                sgRNA_model_char = sgRNA_model_seq[index - sgRNA_gap_count]
                if sgRNA_model_char == query_align[index]:
                    #                 box_value = r"$\bullet$"
                    box_value = r"$\cdot$"
                    box_fill = False
                    box_color = "#FFFFFF"
                else:
                    box_value = query_align[index]
                    box_fill = True
                    box_color = base_color_dict.get(box_value)

                    if not box_color:
                        box_color = "AAAAAA"
            else:
                sgRNA_gap_count += 1
                continue

            box_x_left = box_x_left_vec[index - sgRNA_gap_count]
            box_width = box_x_right_vec[index - sgRNA_gap_count] - box_x_left_vec[index - sgRNA_gap_count]

            # plot align seq box
            patches.append(Rectangle(
                xy=(box_x_left, box_y_btm),
                width=box_width,
                height=panel_box_height,
                fill=box_fill,
                alpha=1,
                linestyle=box_linestyle,
                linewidth=box_linewidth,
                edgecolor=box_edgecolor,
                facecolor=box_color)
            )

            # add sgRNA text
            text_list.append((box_x_left + 0.5 * box_width, box_y_btm + 0.5 * panel_box_height, str(box_value), 16))

            # record index 
            col_index += 1

            # plot name
        for ann_value in ann_value_list:
            box_x_left = box_x_left_vec[col_index]
            box_width = box_x_right_vec[col_index] - box_x_left_vec[col_index]
            try:
                ann_value_str = round(float(ann_value), 4)
            except:
                ann_value_str = str(ann_value)
            text_list.append((box_x_left + 0.5 * box_width, box_y_btm + 0.5 * panel_box_height, ann_value_str, 10))
            col_index += 1

        # make cur_y
        current_y = current_y - panel_box_height - panel_row_space

    # plot remain info
    ax.add_collection(PatchCollection(patches, match_original=True))

    # add text 
    for text_x, text_y, text_info, text_fontsize in text_list:
        plt.text(
            x=text_x,
            y=text_y,
            s=text_info,
            horizontalalignment='center',
            verticalalignment='center',
            fontsize=text_fontsize,
            fontname="Arial"
        )

    fig.savefig(fname=ARGS.out_figure, bbox_inches='tight', dpi=ARGS.out_figure_dpi, format=ARGS.out_figure_format)

# 2022-02-13 by MENG Howard
